"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from collections import namedtuple
from enum import Enum

from von_anchor.a2a.docutil import canon_did, canon_ref


LinkedDataKeySpec = namedtuple('LinkedDataKeySpec', 'ver_type authn_type specifier')


class PublicKeyType(Enum):
    """
    Class encapsulating public key types.
    """

    RSA_SIG_2018 = LinkedDataKeySpec(
        'RsaVerificationKey2018',
        'RsaSignatureAuthentication2018',
        'publicKeyPem')
    ED25519_SIG_2018 = LinkedDataKeySpec(
        'Ed25519VerificationKey2018',
        'Ed25519SignatureAuthentication2018',
        'publicKeyBase58')
    EDDSA_SA_SIG_SECP256K1 = LinkedDataKeySpec(
        'Secp256k1VerificationKey2018',
        'Secp256k1SignatureAuthenticationKey2018',
        'publicKeyHex')

    @staticmethod
    def get(val: str) -> 'PublicKeyType':
        """
        Return enum instance corresponding to input value ('RsaVerificationKey2018' etc.)

        :param val: input value marking public key type
        :return: public key type
        """

        for pktype in PublicKeyType:
            if val in (pktype.ver_type, pktype.authn_type):
                return pktype
        return None

    @property
    def ver_type(self) -> str:
        """
        Return verification type identifier in public key specification.

        :return: verification type
        """

        return self.value.ver_type

    @property
    def authn_type(self) -> str:
        """
        Return authentication type identifier in public key specification.

        :return: authentication type
        """

        return self.value.authn_type

    @property
    def specifier(self) -> str:
        """
        Return value specifier in public key specification.

        :return: value specifier in public key specification
        """

        return self.value.specifier

    def specification(self, val: str) -> str:
        """
        Return specifier and input value for use in public key specification.

        :param val: value of public key
        :return: dict mapping applicable specifier to input value
        """

        return {self.specifier: val}


class PublicKey:
    """
    Public key specification to embed in DID document. Retains DIDs as raw values
    (orientated toward indy-facing operations), everything else as URIs
    (oriented toward W3C-facing operations).
    """

    def __init__(
            self,
            did: str,
            ident: str,
            value: str,
            pk_type: PublicKeyType = None,
            controller: str = None,
            authn: bool = False) -> None:
        """
        Retain key specification particulars. Raise BadIdentifier on any bad input DID.

        :param did: DID of DID document embedding public key
        :param ident: identifier for public key
        :param value: key content, encoded as key specification requires
        :param pk_type: public key type (enum), default ED25519_SIG_2018
        :param controller: controller DID (default DID of DID document)
        :param authn: whether key as has DID authentication privilege (default False)
        """

        self._did = canon_did(did)
        self._id = canon_ref(self._did, ident)
        self._value = value
        self._type = pk_type or PublicKeyType.ED25519_SIG_2018
        self._controller = canon_did(controller) if controller else self._did
        self._authn = authn

    @property
    def did(self) -> str:
        """
        Return DID.

        :return: DID
        """

        return self._did

    @property
    def id(self) -> str:
        """
        Return public key identifier.

        :return: public key identifier
        """

        return self._id

    @property
    def type(self) -> PublicKeyType:
        """
        Return public key type.

        :return: public key type
        """

        return self._type

    @property
    def value(self) -> str:
        """
        Return public key value.

        :return: public key value
        """

        return self._value

    @property
    def controller(self) -> str:
        """
        Return controller DID.

        :return: controller DID
        """

        return self._controller

    @property
    def authn(self) -> bool:
        """
        Return whether public key is marked as having DID authentication privilege.

        :return: whether public key is marked as having DID authentication privilege
        """

        return self._authn

    @authn.setter
    def authn(self, value: bool) -> None:
        """
        Set or clear authentication marker.

        :param value: authentication marker
        """

        self._authn = value

    def to_dict(self):
        """
        Return dict representation of public key to embed in DID document.
        """

        return {
            'id': self.id,
            'type': str(self.type.ver_type),
            'controller': canon_ref(self.did, self.controller),
            **self.type.specification(self.value)
        }

    def __repr__(self):
        """
        Return representation.
        """

        return 'PublicKey({}, {}, {}, {}, {}, {})'.format(
            self.did,
            self.id,
            self.value,
            self.type,
            self.controller,
            self.authn)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import asyncio
import json
import re

from configparser import ConfigParser
from enum import IntEnum
from os.path import expandvars, isfile
from pprint import pformat
from time import time
from typing import Any, Callable, Sequence, Union


def ppjson(dumpit: Any, elide_to: int = None) -> str:
    """
    JSON pretty printer, whether already json-encoded or not

    :param dumpit: object to pretty-print
    :param elide_to: optional maximum length including ellipses ('...')
    :return: json pretty-print
    """

    if elide_to is not None:
        elide_to = max(elide_to, 3) # make room for ellipses '...'
    try:
        rv = json.dumps(json.loads(dumpit) if isinstance(dumpit, str) else dumpit, indent=4)
    except TypeError:
        rv = '{}'.format(pformat(dumpit, indent=4, width=120))
    return rv if elide_to is None or len(rv) <= elide_to else '{}...'.format(rv[0 : elide_to - 3])


def do_wait(coro: Callable) -> Any:
    """
    Perform aynchronous operation; await then return the result.

    :param coro: coroutine to await
    :return: coroutine result
    """

    event_loop = None
    try:
        event_loop = asyncio.get_event_loop()
    except RuntimeError:
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
    return event_loop.run_until_complete(coro)


def inis2dict(ini_paths: Union[str, Sequence[str]]) -> dict:
    """
    Take one or more ini files and return a dict with configuration from all,
    interpolating bash-style variables ${VAR} or ${VAR:-DEFAULT}.

    :param ini_paths: path or paths to .ini files
    """

    var_dflt = r'\${(.*?):-(.*?)}'
    def _interpolate(content):
        rv = expandvars(content)
        while True:
            match = re.search(var_dflt, rv)
            if match is None:
                break
            bash_var = '${{{}}}'.format(match.group(1))
            value = expandvars(bash_var)
            rv = re.sub(var_dflt, match.group(2) if value == bash_var else value, rv, count=1)

        return rv

    parser = ConfigParser()

    for ini in [ini_paths] if isinstance(ini_paths, str) else ini_paths:
        if not isfile(ini):
            raise FileNotFoundError('No such file: {}'.format(ini))
        with open(ini, 'r') as ini_fh:
            ini_text = _interpolate(ini_fh.read())
            parser.read_string(ini_text)

    return {s: dict(parser[s].items()) for s in parser.sections()}


class Stopwatch:
    """
    Stopwatch class for troubleshooting lags.
    """

    def __init__(self, digits: int = None):
        """
        Instantiate and start.

        :param digits: number of fractional decimal digits to retain (default to all) by default
        """

        self._mark = [time()] * 2
        self._digits = digits

    def mark(self, digits: int = None) -> float:
        """
        Return time in seconds since last mark, reset, or construction.

        :param digits: number of fractional decimal digits to retain (default as constructed)
        """

        self._mark[:] = [self._mark[1], time()]
        rv = self._mark[1] - self._mark[0]

        if digits is not None and digits > 0:
            rv = round(rv, digits)
        elif digits == 0 or self._digits == 0:
            rv = int(rv)
        elif self._digits is not None and self._digits > 0:
            rv = round(rv, self._digits)

        return rv

    def reset(self) -> float:
        """
        Reset.
        """

        self._mark = [time()] * 2
        return 0.0


class Ink(IntEnum):
    """
    Class encapsulating ink colours for logging.
    """

    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37

    def __call__(self, message: str) -> str:
        """
        Return input message in colour.

        :return: input message in colour
        """

        return '\033[{}m{}\033[0m'.format(self.value, message)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from os import makedirs
from os.path import expanduser, join
from time import time

from indy import anoncreds, ledger
from indy.error import IndyError

from von_anchor.anchor.base import BaseAnchor
from von_anchor.cache import ArchivableCaches, CRED_DEF_CACHE, REVO_CACHE, SCHEMA_CACHE
from von_anchor.canon import canon
from von_anchor.error import (
    AbsentPool,
    AbsentRevReg,
    AbsentSchema,
    BadIdentifier,
    BadRevStateTime,
    ClosedPool)
from von_anchor.indytween import encode, Predicate, Role
from von_anchor.nodepool import NodePool
from von_anchor.util import cred_def_id2seq_no, ok_cred_def_id, ok_rev_reg_id, ok_schema_id
from von_anchor.validcfg import validate_config
from von_anchor.wallet import Wallet


LOGGER = logging.getLogger(__name__)


class Verifier(BaseAnchor):
    """
    Mixin for anchor acting in the capacity of Verifier. Verifier anchors verify proofs.
    """

    def __init__(self, wallet: Wallet, pool: NodePool = None, **kwargs) -> None:
        """
        Initializer for Verifier anchor. Retain input parameters; do not open wallet.

        :param wallet: wallet for anchor use
        :param pool: pool for anchor use
        :param config: configuration dict for cache archive behaviour; e.g.,

        ::

            {
                'parse-caches-on-open': True,
                'archive-verifier-caches-on-close': {
                    'schema_id': [
                        'R17v42T4pk...:2:tombstone:1.2',
                        '9cHbp54C8n...:2:business:2.0',
                        'Pcq76cx6jE...:2:birth_cert:1.0',
                        ...
                    ],
                    'cred_def_id': [
                        'R17v42T4pk...:3:CL:19:tag',
                        '9cHbp54C8n...:3:CL:37:tag',
                        'Pcq76cx6jE...:3:CL:51:tag',
                        ...
                    ]
                    'rev_reg_id': [
                        'R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0',
                        'R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1',
                        '9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:0',
                        '9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:1',
                        '9cHbp54C8n...:4:9cHbp54C8n...:3:CL:37:tag:CL_ACCUM:2',
                        ...
                    ]
                }
            }

        """

        LOGGER.debug('Verifier.__init__ >>> wallet: %s, pool: %s, kwargs: %s', wallet, pool, kwargs)

        super().__init__(wallet, pool, **kwargs)

        self._config = kwargs.get('config', {})
        validate_config('verifier', self._config)

        self._dir_cache = join(expanduser('~'), '.indy_client', 'cache', self.name)
        makedirs(self._dir_cache, exist_ok=True)

        LOGGER.debug('Verifier.__init__ <<<')

    @staticmethod
    def least_role() -> Role:
        """
        Return the indy-sdk USER (null) role for a verifier, which does not need write access.

        :return: USER role
        """

        LOGGER.debug('AnchorSmith.least_role >>>')

        rv = Role.USER

        LOGGER.debug('AnchorSmith.least_role <<< %s', rv)
        return rv

    @property
    def config(self) -> dict:
        """
        Accessor for configuration dict

        :return: verifier config dict
        """

        return self._config

    @config.setter
    def config(self, value: dict) -> None:
        """
        Set configuration dict

        :param value: configuration dict
        """

        validate_config('verifier', value or {})
        self._config = value or {}

    @property
    def dir_cache(self) -> str:
        """
        Accessor for cache archive directory

        :return: verifier cache archive directory
        """

        return self._dir_cache

    async def _build_rr_state_json(self, rr_id: str, timestamp: int) -> (str, int):
        """
        Build rev reg state json at a given requested timestamp.

        Return rev reg state json and its transaction time on the distributed ledger,
        with upper bound at input timestamp of interest.

        Raise AbsentRevReg if no revocation registry exists on input rev reg id,
        or BadRevStateTime if requested timestamp predates revocation registry creation.

        :param rr_id: rev reg id
        :param timestamp: timestamp of interest (epoch seconds)
        :return: rev reg state json and ledger timestamp (epoch seconds)
        """

        LOGGER.debug('_Verifier._build_rr_state_json >>> rr_id: %s, timestamp: %s', rr_id, timestamp)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('Verifier._build_rr_state_json <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        rr_json = None
        ledger_timestamp = None

        get_rr_req_json = await ledger.build_get_revoc_reg_request(self.did, rr_id, timestamp)
        resp_json = await self._submit(get_rr_req_json)
        resp = json.loads(resp_json)
        if resp.get('result', {}).get('data', None) and resp['result']['data'].get('value', None):
            # timestamp at or beyond rev reg creation, carry on
            try:
                (_, rr_json, ledger_timestamp) = await ledger.parse_get_revoc_reg_response(resp_json)
            except IndyError:  # ledger replied, but there is no such rev reg available
                LOGGER.debug('Verifier._build_rr_state_json <!< no rev reg exists on %s', rr_id)
                raise AbsentRevReg('No rev reg exists on {}'.format(rr_id))
        else:
            LOGGER.debug(
                '_Verifier._build_rr_state_json <!< Rev reg %s created after asked-for time %s',
                rr_id,
                timestamp)
            raise BadRevStateTime('Rev reg {} created after asked-for time {}'.format(rr_id, timestamp))

        rv = (rr_json, ledger_timestamp)
        LOGGER.debug('_Verifier._build_rr_state_json <<< %s', rv)
        return rv

    async def build_proof_req_json(self, cd_id2spec: dict) -> str:
        """
        Build and return indy-sdk proof request for input attributes and non-revocation intervals by cred def id.

        :param cd_id2spec: dict mapping cred def ids to:

            - (optionally) 'attrs': lists of names of attributes of interest (omit for all, empty list or None for none)
            - (optionally) '>=': (pred) inclusive int lower-bounds of interest (omit, empty list, or None for none)
            - (optionally) '>': (pred) exclusive int lower-bounds of interest (omit, empty list, or None for none)
            - (optionally) '<=': (pred) inclusive int upper-bounds of interest (omit, empty list, or None for none)
            - (optionally) '<': (pred) exclusive int upper-bounds of interest (omit, empty list, or None for none)
            - (optionally), 'interval': either
                - (2-tuple) pair of epoch second counts marking 'from' and 'to' timestamps, or
                - | single epoch second count to set 'from' and 'to' the same; default
                  | (now, now) for cred defs supporting revocation or None otherwise; e.g.,

        ::

            {
                'Vx4E82R17q...:3:CL:16:tag': {
                    'attrs': [  # request attrs 'name' and 'favouriteDrink' from this cred def's schema
                        'name',
                        'favouriteDrink'
                    ],
                    '>=': {  # request predicate score>=80 from this cred def
                        'score': 80
                    }
                    '<=': {  # request ranking <=10 from this cred def
                        'ranking': 10
                    }
                    'interval': 1528116008  # same instant for all attrs and preds of corresponding schema
                },
                'R17v42T4pk...:3:CL:19:tag': None,  # request all attrs, no preds, default intervals on all attrs
                'e3vc5K168n...:3:CL:23:tag': {},  # request all attrs, no preds, default intervals on all attrs
                'Z9ccax812j...:3:CL:27:tag': {  # request all attrs, no preds, this interval on all attrs
                    'interval': (1528112408, 1528116008)
                },
                '9cHbp54C8n...:3:CL:37:tag': {  # request no attrs and some predicates; specify interval
                    'attrs': [],  # or equivalently, 'attrs': None
                    '>=': {
                        'employees': '50'  # nicety: implementation converts to int for caller
                    },
                    '>=': {
                        'revenue': '10000000'  # nicety: implementation converts to int for caller
                        'ebidta': 0
                    }
                    'interval': (1528029608, 1528116008)
                },
                '6caBcmLi33...:3:CL:41:tag': {  # all attrs, one pred, default intervals to now on attrs & pred
                    '>': {
                        'regEpoch': 1514782800
                    }
                },
                ...
            }

        :return: indy-sdk proof request json
        """

        LOGGER.debug('Verifier.build_proof_req_json >>> cd_id2spec: %s', cd_id2spec)

        cd_id2schema = {}
        now = int(time())
        rv = {
            'nonce': str(int(time())),
            'name': 'proof_req',
            'version': '0.0',
            'requested_attributes': {},
            'requested_predicates': {}
        }

        for cd_id in cd_id2spec:
            if not ok_cred_def_id(cd_id):
                LOGGER.debug('Verifier.build_proof_req_json <!< Bad cred def id %s', cd_id)
                raise BadIdentifier('Bad cred def id {}'.format(cd_id))

            interval = None
            cred_def = json.loads(await self.get_cred_def(cd_id))
            seq_no = cred_def_id2seq_no(cd_id)
            cd_id2schema[cd_id] = json.loads(await self.get_schema(seq_no))

            if 'revocation' in cred_def['value']:
                fro_to = cd_id2spec[cd_id].get('interval', (now, now)) if cd_id2spec[cd_id] else (now, now)
                interval = {
                    'from': fro_to if isinstance(fro_to, int) else min(fro_to),
                    'to': fro_to if isinstance(fro_to, int) else max(fro_to)
                }

            for attr in (cd_id2spec[cd_id].get('attrs', cd_id2schema[cd_id]['attrNames']) or []
                    if cd_id2spec[cd_id] else cd_id2schema[cd_id]['attrNames']):
                attr_uuid = '{}_{}_uuid'.format(seq_no, canon(attr))
                rv['requested_attributes'][attr_uuid] = {
                    'name': attr,
                    'restrictions': [{
                        'cred_def_id': cd_id
                    }]
                }
                if interval:
                    rv['requested_attributes'][attr_uuid]['non_revoked'] = interval

            for pred in Predicate:
                for attr in (cd_id2spec[cd_id].get(pred.value.math, {}) or {} if cd_id2spec[cd_id] else {}):
                    pred_uuid = '{}_{}_{}_uuid'.format(seq_no, canon(attr), pred.value.fortran)
                    try:
                        rv['requested_predicates'][pred_uuid] = {
                            'name': attr,
                            'p_type': pred.value.math,
                            'p_value': Predicate.to_int(cd_id2spec[cd_id][pred.value.math][attr]),
                            'restrictions': [{
                                'cred_def_id': cd_id
                            }]
                        }
                    except ValueError:
                        LOGGER.info(
                            'cannot build %s predicate on non-int bound %s for %s',
                            pred.value.fortran,
                            cd_id2spec[cd_id][pred.value.math][attr],
                            attr)
                        continue  # int conversion failed - reject candidate
                    if interval:
                        rv['requested_predicates'][pred_uuid]['non_revoked'] = interval

        LOGGER.debug('Verifier.build_proof_req_json <<< %s', json.dumps(rv))
        return json.dumps(rv)

    async def load_cache_for_verification(self, archive: bool = False) -> int:
        """
        Load schema, cred def, revocation caches; optionally archive enough to go
        offline and be able to verify proof on content marked of interest in configuration.

        Return timestamp (epoch seconds) of cache load event, also used as subdirectory
        for cache archives.

        :param archive: True to archive now or False to demur (subclasses may still
            need to augment archivable caches further)
        :return: cache load event timestamp (epoch seconds)
        """

        LOGGER.debug('Verifier.load_cache_for_verification >>> archive: %s', archive)

        rv = int(time())
        for s_id in self.config.get('archive-verifier-caches-on-close', {}).get('schema_id', {}):
            if ok_schema_id(s_id):
                with SCHEMA_CACHE.lock:
                    await self.get_schema(s_id)
            else:
                LOGGER.info('Not archiving schema for specified bad id %s', s_id)
        for cd_id in self.config.get('archive-verifier-caches-on-close', {}).get('cred_def_id', {}):
            if ok_cred_def_id(cd_id):
                with CRED_DEF_CACHE.lock:
                    await self.get_cred_def(cd_id)
            else:
                LOGGER.info('Not archiving cred def for specified bad id %s', cd_id)
        for rr_id in self.config.get('archive-verifier-caches-on-close', {}).get('rev_reg_id', {}):
            if ok_rev_reg_id(rr_id):
                await self.get_rev_reg_def(rr_id)
                with REVO_CACHE.lock:
                    revo_cache_entry = REVO_CACHE.get(rr_id, None)
                    if revo_cache_entry:
                        try:
                            await revo_cache_entry.get_state_json(self._build_rr_state_json, rv, rv)
                        except ClosedPool:
                            LOGGER.warning(
                                'Verifier %s is offline from pool %s, cannot update revo cache reg state for %s to %s',
                                self.name,
                                self.pool.name,
                                rr_id,
                                rv)
                        except AbsentPool:
                            LOGGER.warning(
                                'Verifier %s has no pool, cannot update revo cache reg state for %s to %s',
                                self.name,
                                rr_id,
                                rv)
            else:
                LOGGER.info('Not archiving rev reg for specified bad id %s', rr_id)

        if archive:
            ArchivableCaches.archive(self.dir_cache)
        LOGGER.debug('Verifier.load_cache_for_verification <<< %s', rv)
        return rv

    async def open(self) -> 'Verifier':
        """
        Explicit entry. Perform ancestor opening operations,
        then parse cache from archive if so configured, and
        synchronize revocation registry to tails tree content.

        :return: current object
        """

        LOGGER.debug('Verifier.open >>>')

        await super().open()
        if self.config.get('parse-caches-on-open', False):
            ArchivableCaches.parse(self.dir_cache)

        LOGGER.debug('Verifier.open <<<')
        return self

    async def close(self) -> None:
        """
        Explicit exit. If so configured, populate cache to prove for any creds on schemata,
        cred defs, and rev regs marked of interest in configuration at initialization,
        archive cache, and purge prior cache archives.

        :return: current object
        """

        LOGGER.debug('Verifier.close >>>')

        if self.config.get('archive-verifier-caches-on-close', {}):
            await self.load_cache_for_verification(True)
            ArchivableCaches.purge_archives(self.dir_cache, True)

        await BaseAnchor.close(self)

        LOGGER.debug('Verifier.close <<<')

    @staticmethod
    def check_encoding(proof_req: dict, proof: dict) -> bool:
        """
        Return whether the proof's raw values correspond to their encodings
        as cross-referenced against proof request.

        :param proof request: proof request
        :param proof: corresponding proof to check
        :return: True if OK, False for encoding mismatch
        """

        LOGGER.debug('Verifier.check_encoding <<< proof_req: %s, proof: %s', proof_req, proof)

        cd_id2proof_id = {}  # invert proof['identifiers'] per cd_id
        p_preds = {}  # cd_id and attr to bound
        for idx in range(len(proof['identifiers'])):
            cd_id = proof['identifiers'][idx]['cred_def_id']
            cd_id2proof_id[cd_id] = idx  # since at most 1 cred per cred def
            p_preds[cd_id] = {
                ge_proof['predicate']['attr_name']: ge_proof['predicate']['value']
                for ge_proof in proof['proof']['proofs'][idx]['primary_proof']['ge_proofs']
            }

        for (uuid, req_attr) in proof_req['requested_attributes'].items():  # proof req xref proof per revealed attr
            canon_attr = canon(req_attr['name'])
            proof_ident_idx = cd_id2proof_id[req_attr['restrictions'][0]['cred_def_id']]
            enco = proof['proof']['proofs'][proof_ident_idx]['primary_proof']['eq_proof']['revealed_attrs'].get(
                canon_attr)
            if not enco:
                continue  # requested but declined from revelation in proof: must appear in a predicate
            if enco != proof['requested_proof']['revealed_attrs'][uuid]['encoded']:
                LOGGER.debug('Verifier.check_proof_encoding <<< False')
                return False
            if enco != encode(proof['requested_proof']['revealed_attrs'][uuid]['raw']):
                LOGGER.debug('Verifier.check_proof_encoding <<< False')
                return False

        for (uuid, req_pred) in proof_req['requested_predicates'].items():  # proof req xref proof per pred
            canon_attr = canon(req_pred['name'])
            if p_preds[req_pred['restrictions'][0]['cred_def_id']].get(canon_attr) != req_pred['p_value']:
                LOGGER.debug('Verifier.check_proof_encoding <<< False')
                return False

        LOGGER.debug('Verifier.check_proof_encoding <<< True')
        return True

    async def verify_proof(self, proof_req: dict, proof: dict) -> str:
        """
        Verify proof as Verifier. Raise AbsentRevReg if a proof cites a revocation registry
        that does not exist on the distributed ledger.

        :param proof_req: proof request as Verifier creates, as per proof_req_json above
        :param proof: proof as HolderProver creates
        :return: json encoded True if proof is valid; False if not
        """

        LOGGER.debug('Verifier.verify_proof >>> proof_req: %s, proof: %s', proof_req, proof)

        if not Verifier.check_encoding(proof_req, proof):
            LOGGER.info(
                'Proof encoding does not cross-reference with proof request %s: failing verification',
                proof_req.get('nonce', '(missing nonce)'))
            LOGGER.debug('Verifier.verify_proof <<< "False"')
            return json.dumps(False)

        async def _set_schema(s_id: str) -> None:
            nonlocal s_id2schema
            if not ok_schema_id(s_id):
                LOGGER.debug('Verifier.verify_proof <!< Bad schema id %s', s_id)
                raise BadIdentifier('Bad schema id {}'.format(s_id))
            if s_id not in s_id2schema:
                schema = json.loads(await self.get_schema(s_id))  # add to cache en passant
                if not schema:
                    LOGGER.debug(
                        'Verifier.verify_proof <!< absent schema %s, proof req may be for another ledger',
                        s_id)
                    raise AbsentSchema('Absent schema {}, proof req may be for another ledger'.format(s_id))
                s_id2schema[s_id] = schema

        async def _set_cred_def(cd_id: str) -> None:
            nonlocal cd_id2cred_def
            if not ok_cred_def_id(cd_id):
                LOGGER.debug('Verifier.verify_proof <!< Bad cred def id %s', cd_id)
                raise BadIdentifier('Bad cred def id {}'.format(cd_id))
            if cd_id not in cd_id2cred_def:
                cd_id2cred_def[cd_id] = json.loads(await self.get_cred_def(cd_id))  # add to cache en passant

        async def _set_rev_reg_def(rr_id: str) -> bool:
            """
            Return true to continue to timestamp setting, false to short-circuit
            """
            nonlocal rr_id2rr_def
            if not rr_id:
                return False
            if not ok_rev_reg_id(rr_id):
                LOGGER.debug('Verifier.verify_proof <!< Bad rev reg id %s', rr_id)
                raise BadIdentifier('Bad rev reg id {}'.format(rr_id))
            if rr_id not in rr_id2rr_def:
                rr_id2rr_def[rr_id] = json.loads(await self.get_rev_reg_def(rr_id))
            return True

        async def _set_timestamp(rr_id: str, timestamp: int) -> None:
            nonlocal rr_id2rr
            with REVO_CACHE.lock:
                revo_cache_entry = REVO_CACHE.get(rr_id, None)
                (rr_json, _) = await revo_cache_entry.get_state_json(self._build_rr_state_json, timestamp, timestamp)
                if rr_id not in rr_id2rr:
                    rr_id2rr[rr_id] = {}
                rr_id2rr[rr_id][timestamp] = json.loads(rr_json)

        s_id2schema = {}
        cd_id2cred_def = {}
        rr_id2rr_def = {}
        rr_id2rr = {}
        proof_ids = proof['identifiers']

        for proof_id in proof_ids:
            await _set_schema(proof_id['schema_id'])
            await _set_cred_def(proof_id['cred_def_id'])
            rr_id = proof_id['rev_reg_id']
            if await _set_rev_reg_def(rr_id):
                await _set_timestamp(rr_id, proof_id['timestamp'])

        rv = json.dumps(await anoncreds.verifier_verify_proof(
            json.dumps(proof_req),
            json.dumps(proof),
            json.dumps(s_id2schema),
            json.dumps(cd_id2cred_def),
            json.dumps(rr_id2rr_def),
            json.dumps(rr_id2rr)))

        LOGGER.debug('Verifier.verify_proof <<< %s', rv)
        return rv



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from von_anchor.canon import raw, canon_pairwise_tag
from von_anchor.error import BadRecord
from von_anchor.wallet.record import StorageRecord


class KeyInfo:
    """
    Bundle for verification key and metadata.
    """

    def __init__(self, verkey: str, metadata: dict = None) -> None:
        """
        Initialize verification key, metadata.

        :param verkey: verification key to store
        :param metadata: metadata associated with current DID
        """

        self._verkey = verkey
        self._metadata = metadata or {}  # store trivial metadata as empty (for iteration), return as None

    @property
    def verkey(self) -> str:
        """
        Accessor for verification key

        :return: verification key
        """

        return self._verkey

    @property
    def metadata(self) -> dict:
        """
        Accessor for metadata

        :return: metadata
        """

        return self._metadata or None  # store trivial metadata as empty (for iteration), return as None

    @metadata.setter
    def metadata(self, value: dict) -> None:
        """
        Accessor for metadata

        :param value: metadata dict
        """

        self._metadata = value or {}  # store trivial metadata as empty (for iteration), return as None

    def __eq__(self, other: 'KeyInfo') -> bool:
        """
        Equivalence operator. Two KeyInfos are equivalent when their attributes are.

        :param other: KeyInfo to test for equivalence
        :return: whether KeyInfos are equivalent
        """

        return self.verkey == other.verkey and self.metadata == other.metadata

    def __repr__(self) -> str:
        """
        Return representation.

        :return: string representation evaluating to construction call
        """

        return 'KeyInfo({}, {})'.format(self.verkey, self.metadata)

class DIDInfo:
    """
    Bundle for DID, verification key, and metadata.
    """

    def __init__(self, did: str, verkey: str, metadata: dict = None) -> None:
        """
        Initialize DID, verification key, metadata.

        :param did: DID to store
        :param verkey: verification key to store
        :param metadata: metadata associated with current DID
        """

        self._did = did
        self._verkey = verkey
        self._metadata = metadata or {}  # store trivial metadata as empty (for iteration), return as None

    @property
    def did(self) -> str:
        """
        Accessor for DID

        :return: DID
        """

        return self._did

    @property
    def verkey(self) -> str:
        """
        Accessor for verification key

        :return: verification key
        """

        return self._verkey

    @property
    def metadata(self) -> dict:
        """
        Accessor for metadata

        :return: metadata
        """

        return self._metadata or None  # store trivial metadata as empty (for iteration), return as None

    @metadata.setter
    def metadata(self, value: dict) -> None:
        """
        Accessor for metadata

        :param value: metadata dict
        """

        self._metadata = value or {}  # store trivial metadata as empty (for iteration), return as None

    def __eq__(self, other: 'DIDInfo') -> bool:
        """
        Equivalence operator. Two DIDInfos are equivalent when their attributes are.

        :param other: DIDInfo to test for equivalence
        :return: whether DIDInfos are equivalent
        """

        return self.did == other.did and self.verkey == other.verkey and self.metadata == other.metadata

    def __repr__(self) -> str:
        """
        Return representation.

        :return: string representation evaluating to construction call
        """

        return 'DIDInfo({}, {}, {})'.format(self.did, self.verkey, self.metadata)


class EndpointInfo:
    """
    Bundle for endpoint and (transport) verification key.
    """

    def __init__(self, endpoint: str, verkey: str) -> None:
        """
        Initialize endpoint, verification key.

        :param endpoint: endpoint to store ('<ip-address>:<port>')
        :param verkey: verification key to store
        """

        self._endpoint = endpoint
        self._verkey = verkey

    @property
    def endpoint(self) -> str:
        """
        Accessor for endpoint

        :return: endpoint
        """

        return self._endpoint

    @property
    def ip_addr(self) -> str:
        """
        Accessor for endpoint IP address

        :return: endpoint IP address
        """

        return self._endpoint.split(':')[0]

    @property
    def port(self) -> int:
        """
        Accessor for endpoint port

        :return: endpoint port
        """

        return int(self._endpoint.split(':')[-1])

    @property
    def verkey(self) -> str:
        """
        Accessor for verification key

        :return: verification key
        """

        return self._verkey

    def __eq__(self, other: 'EndpointInfo') -> bool:
        """
        Equivalence operator. Two EndpointInfos are equivalent when their attributes are.

        :param other: EndpointInfo to test for equivalence
        :return: whether EndpointInfos are equivalent
        """

        return self.endpoint == other.endpoint and self.verkey == other.verkey

    def __repr__(self) -> str:
        """
        Return representation.

        :return: string representation evaluating to construction call
        """

        return 'EndpointInfo({}, {})'.format(self.endpoint, self.verkey)


class PairwiseInfo:
    """
    Bundle for pairwise DID relation: DIDs, verification keys, and metadata.
    """

    def __init__(self, their_did: str, their_verkey: str, my_did: str, my_verkey: str, metadata: dict = None) -> None:
        """
        Initialize DIDs, verification keys, metadata.

        :param their_did: remote DID to store
        :param their_verkey: remote verification key to store
        :param my_did: local DID to store
        :param my_verkey: local verification key to store
        :param metadata: metadata associated with pairwise DID relationship
        """

        self._their_did = their_did
        self._their_verkey = their_verkey
        self._my_did = my_did
        self._my_verkey = my_verkey
        self._metadata = metadata

    @property
    def their_did(self) -> str:
        """
        Accessor for remote DID

        :return: remote DID
        """

        return self._their_did

    @property
    def their_verkey(self) -> str:
        """
        Accessor for remote verification key

        :return: remote verification key
        """

        return self._their_verkey

    @property
    def my_did(self) -> str:
        """
        Accessor for local DID

        :return: local DID
        """

        return self._my_did

    @property
    def my_verkey(self) -> str:
        """
        Accessor for local verification key

        :return: local verification key
        """

        return self._my_verkey

    @property
    def metadata(self) -> dict:
        """
        Accessor for metadata

        :return: metadata
        """

        return self._metadata

    @metadata.setter
    def metadata(self, value: dict) -> None:
        """
        Accessor for metadata

        :param value: metadata dict
        """

        self._metadata = value

    def __eq__(self, other: 'PairwiseInfo') -> bool:
        """
        Equivalence operator. Two PairwiseInfos are equivalent when their attributes are.

        :param other: PairwiseInfo to test for equivalence
        :return: whether PairwiseInfos are equivalent
        """

        return (
            self.their_did == other.their_did and
            self.their_verkey == other.their_verkey and
            self.my_did == other.my_did and
            self.my_verkey == other.my_verkey and
            self.metadata == other.metadata)

    def __repr__(self) -> str:
        """
        Return representation.

        :return: string representation evaluating to construction call
        """

        return 'PairwiseInfo({}, {}, {}, {}, {})'.format(
            self.their_did,
            self.their_verkey,
            self.my_did,
            self.my_verkey,
            self.metadata)


def pairwise_info2tags(pairwise: PairwiseInfo) -> dict:
    """
    Given pairwise info with metadata mapping tags to values, return corresponding
    indy-sdk non_secrets record tags dict to store same in wallet (via non_secrets)
    unencrypted (for WQL search options).  Canonicalize metadata values to strings via
    raw() for WQL fitness.

    Raise BadRecord if metadata does not coerce into non_secrets API tags spec of {str:str}.

    :param pairwise: pairwise info with metadata dict mapping tags to values
    :return: corresponding non_secrets tags dict marked for unencrypted storage
    """

    rv = {
        canon_pairwise_tag(tag): raw(pairwise.metadata[tag]) for tag in pairwise.metadata or {}
    }
    rv['~their_did'] = pairwise.their_did
    rv['~their_verkey'] = pairwise.their_verkey
    rv['~my_did'] = pairwise.my_did
    rv['~my_verkey'] = pairwise.my_verkey

    if not StorageRecord.ok_tags(rv):
        raise BadRecord('Pairwise metadata {} must map strings to strings'.format(rv))

    return rv


def storage_record2pairwise_info(storec: StorageRecord) -> PairwiseInfo:
    """
    Given indy-sdk non_secrets implementation of pairwise storage record dict, return corresponding PairwiseInfo.

    :param storec: (non-secret) storage record to convert to PairwiseInfo
    :return: PairwiseInfo on record DIDs, verkeys, metadata
    """

    return PairwiseInfo(
        storec.id,  # = their did
        storec.value,  # = their verkey
        storec.tags['~my_did'],
        storec.tags['~my_verkey'],
        {
            tag[tag.startswith('~'):]: storec.tags[tag] for tag in (storec.tags or {})  # strip any leading '~'
        })



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from typing import List, Sequence, Union

from von_anchor.a2a.docutil import canon_did, canon_ref, resource
from von_anchor.a2a.publickey import PublicKey, PublicKeyType
from von_anchor.a2a.service import Service
from von_anchor.error import AbsentDIDDocItem, BadDIDDocItem, BadIdentifier
from von_anchor.util import ok_did


LOGGER = logging.getLogger(__name__)


class DIDDoc:
    """
    DID document, grouping a DID with verification keys and services.
    Retains DIDs as raw values (orientated toward indy-facing operations),
    everything else as URIs (oriented toward W3C-facing operations).
    """

    CONTEXT = 'https://w3id.org/did/v1'

    def __init__(self, did: str = None) -> None:
        """
        Initializer. Retain DID ('id' in DIDDoc context); initialize verification keys and services to empty lists.
        Raise BadIdentifier for bad input DID.

        :param did: DID for current DIDdoc
        """

        self._did = canon_did(did) if did else None  # allow specification post-hoc
        self._pubkey = {}
        self._service = {}

    @property
    def did(self) -> str:
        """
        Accessor for DID.
        """

        return self._did

    @did.setter
    def did(self, value: str) -> None:
        """
        Set DID ('id' in DIDDoc context). Raise BadIdentifier for bad input DID.

        :param value: DID
        """

        self._did = canon_did(value) if value else None

    @property
    def pubkey(self) -> dict:
        """
        Accessor for public keys by identifier.
        """

        return self._pubkey

    @property
    def authnkey(self) -> dict:
        """
        Accessor for public keys marked as authentication keys, by identifier.
        """

        return {k: self._pubkey[k] for k in self._pubkey if self._pubkey[k].authn}

    @property
    def service(self) -> dict:
        """
        Accessor for services by identifier.
        """

        return self._service

    def set(self, item: Union[Service, PublicKey]) -> 'DIDDoc':
        """
        Add or replace service or public key; return current DIDDoc.

        Raise BadDIDDocItem if input item is neither service nor public key.

        :param item: service or public key to set
        :return: current DIDDoc
        """

        if isinstance(item, Service):
            self.service[item.id] = item
        elif isinstance(item, PublicKey):
            self.pubkey[item.id] = item
        else:
            raise BadDIDDocItem('Cannot add item {} to DIDDoc on DID {}'.format(item, self.did))

    def serialize(self) -> str:
        """
        Dump current object to a JSON-compatible dictionary.

        :return: dict representation of current DIDDoc
        """

        return {
            '@context': DIDDoc.CONTEXT,
            'id': canon_ref(self.did, self.did),
            'publicKey': [pubkey.to_dict() for pubkey in self.pubkey.values()],
            'authentication': [{
                'type': pubkey.type.authn_type,
                'publicKey': canon_ref(self.did, pubkey.id)
            } for pubkey in self.pubkey.values() if pubkey.authn],
            'service': [service.to_dict() for service in self.service.values()]
        }

    def to_json(self) -> str:
        """
        Dump current object as json (JSON-LD).

        :return json representation of current DIDDoc
        """

        return json.dumps(self.serialize())

    def add_service_pubkeys(self, service: dict, tags: Union[Sequence[str], str]) -> List[PublicKey]:
        """
        Add public keys specified in service. Return public keys so discovered.

        Raise AbsentDIDDocItem for public key reference not present in DID document.

        :param service: service from DID document
        :param tags: potential tags marking public keys of type of interest - the standard is still coalescing
        :return: list of public keys that service specification in DID document identifies.
        """

        rv = []
        for tag in [tags] if isinstance(tags, str) else list(tags):

            for svc_key in service.get(tag, {}):
                canon_key = canon_ref(self.did, svc_key)
                pubkey = None

                if '#' in svc_key:
                    if canon_key in self.pubkey:
                        pubkey = self.pubkey[canon_key]
                    else:  # service key refers to another DID doc
                        LOGGER.debug(
                            'DIDDoc.add_service_pubkeys <!< DID document %s has no public key %s',
                            self.did,
                            svc_key)
                        raise AbsentDIDDocItem('DID document {} has no public key {}'.format(self.did, svc_key))
                else:
                    for existing_pubkey in self.pubkey.values():
                        if existing_pubkey.value == svc_key:
                            pubkey = existing_pubkey
                            break
                    else:
                        pubkey = PublicKey(
                            self.did,
                            ident=svc_key[-9:-1],  # industrial-grade uniqueness
                            value=svc_key)
                        self._pubkey[pubkey.id] = pubkey

                if pubkey and pubkey not in rv:  # perverse case: could specify same key multiple ways; append once
                    rv.append(pubkey)

        return rv

    @classmethod
    def deserialize(cls, did_doc: dict) -> 'DIDDoc':
        """
        Construct DIDDoc object from dict representation.

        Raise BadIdentifier for bad DID. Raise AbsentDIDDocItem for missing mandatory item.

        :param did_doc: DIDDoc dict reprentation.
        :return: DIDDoc from input json.
        """

        rv = None
        if 'id' in did_doc:
            rv = DIDDoc(did_doc['id'])
        else:  # heuristic: get DID to serve as DID document identifier from first OK-looking public key
            for section in ('publicKey', 'authentication'):
                if rv is None and section in did_doc:
                    for key_spec in did_doc[section]:
                        try:
                            pubkey_did = canon_did(resource(key_spec.get('id', '')))
                            if ok_did(pubkey_did):
                                rv = DIDDoc(pubkey_did)
                                break
                        except BadIdentifier:  # no identifier here, move on to next candidate
                            break
            if rv is None:
                LOGGER.debug('DIDDoc.deserialize <!< no identifier in DID document')
                raise AbsentDIDDocItem('No identifier in DID document')

        for pubkey in did_doc.get('publicKey', {}):  # include all public keys, authentication pubkeys by reference
            pubkey_type = PublicKeyType.get(pubkey['type'])
            authn = any(
                canon_ref(rv.did, ak.get('publicKey', '')) == canon_ref(rv.did, pubkey['id'])
                for ak in did_doc.get('authentication', {}) if isinstance(ak.get('publicKey', None), str))
            key = PublicKey(  # initialization canonicalizes id
                rv.did,
                pubkey['id'],
                pubkey[pubkey_type.specifier],
                pubkey_type,
                canon_did(pubkey['controller']),
                authn)
            rv.pubkey[key.id] = key

        for akey in did_doc.get('authentication', {}):  # include embedded authentication keys
            if 'publicKey' not in akey:  # not yet got it with public keys
                pubkey_type = PublicKeyType.get(akey['type'])
                key = PublicKey(  # initialization canonicalized id
                    rv.did,
                    akey['id'],
                    akey[pubkey_type.specifier],
                    pubkey_type,
                    canon_did(akey['controller']),
                    True)
                rv.pubkey[key.id] = key

        for service in did_doc.get('service', {}):
            endpoint = service['serviceEndpoint']
            svc = Service(  # initialization canonicalizes id
                rv.did,
                service.get('id', canon_ref(rv.did, 'assigned-service-{}'.format(len(rv.service)), ';')),
                service['type'],
                rv.add_service_pubkeys(service, 'recipientKeys'),
                rv.add_service_pubkeys(service, ['mediatorKeys', 'routingKeys']),
                canon_ref(rv.did, endpoint, ';') if ';' in endpoint else endpoint,
                service.get('priority', None))
            rv.service[svc.id] = svc

        return rv

    @classmethod
    def from_json(cls, did_doc_json: str) -> 'DIDDoc':
        """
        Construct DIDDoc object from json representation.

        :param did_doc_json: DIDDoc json reprentation.
        :return: DIDDoc from input json.
        """

        return cls.deserialize(json.loads(did_doc_json))

    def __str__(self) -> str:
        """
        Return string representation for abbreviated display.

        :return: string representation
        """

        return 'DIDDoc({})'.format(self.did)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from collections import namedtuple
from enum import Enum
from hashlib import sha256
from typing import Any, Union

from von_anchor.canon import raw


SchemaKey = namedtuple('SchemaKey', 'origin_did name version')
Relation = namedtuple('Relation', 'fortran wql math yes no')

I32_BOUND = 2**31


def encode(orig: Any) -> str:
    """
    Encode credential attribute value, purely stringifying any int32 and leaving numeric int32 strings alone,
    but mapping any other input to a stringified 256-bit (but not 32-bit) integer. Predicates in indy-sdk operate
    on int32 values properly only when their encoded values match their raw values.

    :param orig: original value to encode
    :return: encoded value
    """

    if isinstance(orig, int) and -I32_BOUND <= orig < I32_BOUND:
        return str(int(orig))  # python bools are ints

    try:
        i32orig = int(str(orig))  # don't encode floats as ints
        if -I32_BOUND <= i32orig < I32_BOUND:
            return str(i32orig)
    except (ValueError, TypeError):
        pass

    return str(int.from_bytes(sha256(raw(orig).encode()).digest(), 'big'))  # OK for indy: sha256 changes all str(i32)s


def cred_attr_value(orig: Any) -> dict:
    """
    Given a value, return corresponding credential attribute value dict for indy-sdk processing.

    :param orig: original attribute value of any stringifiable type
    :return: dict on 'raw' and 'encoded' keys for indy-sdk processing
    """
    return {'raw': raw(orig), 'encoded': encode(orig)}


class Predicate(Enum):
    """
    Enum for predicate types that indy-sdk supports.
    """

    LT = Relation(
        'LT',
        '$lt',
        '<',
        lambda x, y: Predicate.to_int(x) < Predicate.to_int(y),
        lambda x, y: Predicate.to_int(x) >= Predicate.to_int(y))
    LE = Relation(
        'LE',
        '$lte',
        '<=',
        lambda x, y: Predicate.to_int(x) <= Predicate.to_int(y),
        lambda x, y: Predicate.to_int(x) > Predicate.to_int(y))
    GE = Relation(
        'GE',
        '$gte',
        '>=',
        lambda x, y: Predicate.to_int(x) >= Predicate.to_int(y),
        lambda x, y: Predicate.to_int(x) < Predicate.to_int(y))
    GT = Relation(
        'GT',
        '$gt',
        '>',
        lambda x, y: Predicate.to_int(x) > Predicate.to_int(y),
        lambda x, y: Predicate.to_int(x) <= Predicate.to_int(y))

    @staticmethod
    def get(relation: str) -> 'Predicate':
        """
        Return enum instance corresponding to input relation string
        """

        for pred in Predicate:
            if relation.upper() in (pred.value.fortran, pred.value.wql.upper(), pred.value.math):
                return pred
        return None

    @staticmethod
    def to_int(value: Any) -> int:
        """
        Cast a value as its equivalent int for indy predicate argument. Raise ValueError for any input but
        int, stringified int, or boolean.

        :param value: value to coerce.
        """

        if isinstance(value, (bool, int)):
            return int(value)
        return int(str(value))  # kick out floats


class Role(Enum):
    """
    Enum for indy roles.
    """

    STEWARD = (2,)
    TRUSTEE = (0,)
    TRUST_ANCHOR = (101,)
    USER = (None, '')  # reading from config, default empty specifier '' or None to USER
    ROLE_REMOVE = ('',)  # but indy-sdk uses '' to identify a role in reset

    @staticmethod
    def get(token: Union[str, int] = None) -> 'Role':
        """
        Return enum instance corresponding to input token.

        :param token: token identifying role to indy-sdk: 'STEWARD', 'TRUSTEE', 'TRUST_ANCHOR', '' or None
        :return: enum instance corresponding to input token
        """

        if token is None:
            return Role.USER

        for role in Role:
            if role == Role.ROLE_REMOVE:
                continue  # ROLE_REMOVE is not a sensible role to parse from any configuration
            if isinstance(token, int) and token in role.value:
                return role
            if str(token).upper() == role.name or token in (str(v) for v in role.value):  # could be numeric string
                return role

        return None

    def to_indy_num_str(self) -> str:
        """
        Return (typically, numeric) string value that indy-sdk associates with role.

        :return: associated string value (None for self-sovereign user
            having no additional write privileges to ledger, '' for role in reset)
        """

        return self.value[0]

    def token(self) -> str:
        """
        Return token identifying role to indy-sdk.

        :return: token: 'STEWARD', 'TRUSTEE', 'TRUST_ANCHOR', or None (for USER)
        """

        return self.value[0] if self in (Role.USER, Role.ROLE_REMOVE) else self.name



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from enum import IntEnum


class ErrorCode(IntEnum):
    """
    Error codes particular to von_anchor operation.
    """

    Success = 0

    # Errors to do with von_anchor operation
    CredentialFocus = 1003
    AbsentLinkSecret = 1005
    CorruptWallet = 1006
    AbsentSchema = 1007
    AbsentCredDef = 1008
    AbsentTails = 1009
    CorruptTails = 1010
    BadRevocation = 1011
    BadLedgerTxn = 1012
    BadRevStateTime = 1013
    AbsentInterval = 1014
    AbsentRevReg = 1015
    BadIdentifier = 1016
    AbsentProcess = 1017
    BadKey = 1018
    BadRole = 1019
    BadAttribute = 1020
    AbsentNym = 1021

    # Errors to do with wallet operation
    AbsentWallet = 3000
    BadWalletQuery = 3001
    AbsentCred = 3002
    ExtantWallet = 3004
    WalletState = 3005
    ExtantRecord = 3006
    AbsentRecord = 3007
    AbsentMessage = 3008
    BadRecord = 3009
    BadAccess = 3010
    BadSearch = 3011

    # Errors to do with node pool management and operation
    ClosedPool = 4000
    AbsentPool = 4002
    ExtantPool = 4003

    # Errors to do with caching
    CacheIndex = 5000

    # Errors in DID Doc processing
    AbsentDIDDocItem = 6000
    BadDIDDocItem = 6001

    # JSON validation
    JSONValidation = 9000


class VonAnchorError(Exception):
    """
    Error class for von_anchor operation.
    """

    def __init__(self, error_code: ErrorCode, message: str):
        """
        Initialize on code and message.

        :param error_code: error code
        :param message: error message
        """

        super().__init__(message)
        self.error_code = error_code
        self.message = message

    def __str__(self):
        """
        String representation of error.
        """

        return '({}) {}'.format(self.error_code, self.message)

class CredentialFocus(VonAnchorError):
    """
    Attempt to prove credential in specification that resolve to no claims,
    or to multiple claims for any single claim definition.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.CredentialFocus, message)


class CorruptWallet(VonAnchorError):
    """
    Anchor wallet is inconsistent with ledger.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.CorruptWallet, message)


class AbsentSchema(VonAnchorError):
    """
    (HolderProver) anchor attempting operation requiring unavailable schema.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentSchema, message)


class AbsentCredDef(VonAnchorError):
    """
    (Issuer or HolderProver) anchor attempting operation requiring unavailable claim definition.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentCredDef, message)


class AbsentTails(VonAnchorError):
    """
    (Issuer or HolderProver) anchor attempting to open nonexistent tails file.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentTails, message)


class CorruptTails(VonAnchorError):
    """
    (Issuer or HolderProver) anchor attempting to sync tails dir from distributed ledger not having
    corresponding revocation registry.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.CorruptTails, message)


class BadRevocation(VonAnchorError):
    """
    Issuer anchor attempting to perform illegitimate revocation
    (another issuer issued credential, credential revoked already, etc.).
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadRevocation, message)


class BadLedgerTxn(VonAnchorError):
    """
    Ledger rejected transaction.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadLedgerTxn, message)


class BadRevStateTime(VonAnchorError):
    """
    Proof request includes revocation state timestamp for credential before its revocation registry creation,
    or in the future.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadRevStateTime, message)


class AbsentLinkSecret(VonAnchorError):
    """
    (HolderProver) anchor attempting operation requiring absent link secret.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentLinkSecret, message)


class AbsentInterval(VonAnchorError):
    """
    (HolderProver) anchor attempting to create proof on credentials missing
    a non-revocation interval for credential definition that supports revocation.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentInterval, message)


class AbsentRevReg(VonAnchorError):
    """
    Anchor attempting to create revocation registry state but
    revocation registry is not present on the ledger.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentRevReg, message)


class BadIdentifier(VonAnchorError):
    """
    Encountered incorrectly formatted:
      - distributed identifier (DID),
      - schema identifier,
      - credential definition identifier,
      - revocation registry identifier, or
      - tails hash.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadIdentifier, message)


class AbsentProcess(VonAnchorError):
    """
    Required external process not running.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentProcess, message)


class BadKey(VonAnchorError):
    """
    Cryptographic operation revealed bad key.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadKey, message)


class BadRole(VonAnchorError):
    """
    Invalid specifier for indy-sdk role.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadRole, message)


class BadAttribute(VonAnchorError):
    """
    Invalid specifier for indy-sdk role.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadAttribute, message)


class AbsentNym(VonAnchorError):
    """
    Ledger has no such cryptonym.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentNym, message)


class AbsentWallet(VonAnchorError):
    """
    Wallet has not been created (within indy-sdk).
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentWallet, message)


class BadWalletQuery(VonAnchorError):
    """
    Wallet query is badly formed.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadWalletQuery, message)


class AbsentCred(VonAnchorError):
    """
    No such credential.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentCred, message)


class ExtantWallet(VonAnchorError):
    """
    Wallet already exists.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.ExtantWallet, message)


class WalletState(VonAnchorError):
    """
    Wallet needs to be open.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.WalletState, message)


class ExtantRecord(VonAnchorError):
    """
    Wallet record already exists.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.ExtantRecord, message)


class AbsentRecord(VonAnchorError):
    """
    Wallet record does not exist.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentRecord, message)


class AbsentMessage(VonAnchorError):
    """
    Absent target for cryptographic operation.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentMessage, message)


class BadRecord(VonAnchorError):
    """
    Record as specified is not valid for wallet storage.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadRecord, message)


class BadAccess(VonAnchorError):
    """
    Wallet access credentials value is incorrect.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadAccess, message)


class BadSearch(VonAnchorError):
    """
    Search operation failed.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadSearch, message)


class ClosedPool(VonAnchorError):
    """
    Pool needs to be open.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.ClosedPool, message)


class AbsentPool(VonAnchorError):
    """
    Pool required for operation but none is present.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentPool, message)


class ExtantPool(VonAnchorError):
    """
    Pool already exists.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.ExtantPool, message)


class CacheIndex(VonAnchorError):
    """
    Indexation error on a cache.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.CacheIndex, message)


class AbsentDIDDocItem(VonAnchorError):
    """
    Absent item in DID document.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.AbsentDIDDocItem, message)


class BadDIDDocItem(VonAnchorError):
    """
    Bad item in DID document processing.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.BadDIDDocItem, message)


class JSONValidation(VonAnchorError):
    """
    Exception in JSON validation.
    """

    def __init__(self, message: str):
        """
        Initialize on message.

        :param message: error message
        """

        super().__init__(ErrorCode.JSONValidation, message)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import argparse
import asyncio
import json
import logging

from enum import Enum
from os import getpid, kill, listdir, makedirs, remove
from os.path import basename, dirname, expanduser, isdir, isfile, join, realpath
from shutil import rmtree
from subprocess import Popen
from sys import path as sys_path

from indy import anoncreds, blob_storage

DIR_VON_ANCHOR = dirname(dirname(dirname(realpath(__file__))))
if DIR_VON_ANCHOR not in sys_path:
    sys_path.append(DIR_VON_ANCHOR)

from von_anchor.anchor.base import BaseAnchor
from von_anchor.error import AbsentProcess, BadIdentifier, WalletState
from von_anchor.tails import Tails
from von_anchor.util import ok_rev_reg_id, rev_reg_id2cred_def_id, rev_reg_id2cred_def_id_tag
from von_anchor.wallet import Wallet, WalletManager


LOGGER = logging.getLogger(__name__)


_STATE = Enum('_STATE', 'ABSENT RUNNING STOPPING')


class RevRegBuilder(BaseAnchor):
    """
    Issuer alter ego to build revocation registries in parallel to regular Issuer processing.

    Once the indy-sdk can operate from a subprocess, this logic can go into the Issuer itself,
    which can spawn it as a subprocess. Until then, to build revocation registries in parallel,
    whatever process creates an Issuer must create its corresponding RevRegBuilder separately.
    """

    DIR_TAILS = join(expanduser('~'), '.indy_client', 'tails')
    DIR_TAILS_HOPPER = join(DIR_TAILS, '.hopper')

    def __init__(self, wallet: Wallet, **kwargs) -> None:
        """
        Initializer for RevRegBuilder anchor. Retain input parameters; do not open wallet nor tails writer.

        :param wallet: wallet for anchor use
        :param rrbx: whether revocation registry builder is an external process from the Issuer
            (default False for backward-compatible behaviour)
        """

        LOGGER.debug('RevRegBuilder.__init__ >>> wallet: %s, kwargs: %s', wallet, kwargs)

        super().__init__(wallet, None, **kwargs)
        self._external = kwargs.get('rrbx', False)
        self._dir_tails_sentinel = RevRegBuilder.dir_tails_sentinel(wallet.name) if self.external else None

        if self.external:
            makedirs(self.dir_tails_hopper, exist_ok=True)  # self is RevRegBuilder or descendant: spawn rrbx proc
            makedirs(self._dir_tails_sentinel, exist_ok=True)

            rrb_state = RevRegBuilder._get_state(wallet.name)

            if rrb_state == _STATE.STOPPING:
                try:  # cancel the stop order
                    remove(join(RevRegBuilder.dir_tails_sentinel(wallet.name), '.stop'))
                except FileNotFoundError:
                    pass  # too late, it's gone
                else:
                    rrb_state = _STATE.ABSENT

            if rrb_state == _STATE.ABSENT:  # run it
                start_data_json = self._start_data_json()
                with open(join(RevRegBuilder.dir_tails_sentinel(wallet.name), '.start'), 'w') as fh_start:
                    print(start_data_json, file=fh_start)
                rrb_proc = Popen([
                    'python',
                    realpath(__file__),
                    '-n',
                    wallet.name])
                if rrb_proc and rrb_proc.pid:
                    LOGGER.info(
                        '%s %s spawned pid %s to run external revocation registry builder',
                        type(self).__name__,
                        wallet.name,
                        rrb_proc.pid)
                else:
                    LOGGER.debug('%s %s could not spawn rev reg builder', type(self).__name__, wallet.name)
                    raise AbsentProcess('RevRegBuilder.__init__ <!< {} {} could not spawn rev reg builder'.format(
                        type(self).__name__,
                        wallet.name))

        LOGGER.debug('RevRegBuilder.__init__ <<<')

    @property
    def external(self) -> bool:
        """
        Accessor for revocation registry builder posture: external (True) or internal (False)

        :return: whether revocation registry builder posture is external
        """

        return self._external

    @property
    def dir_tails(self) -> str:
        """
        Accessor for root of tails directory

        :return: tails directory
        """

        return RevRegBuilder.DIR_TAILS

    @property
    def dir_tails_hopper(self) -> str:
        """
        Accessor for tails hopper directory

        :return: tails hopper directory
        """

        return RevRegBuilder.DIR_TAILS_HOPPER

    def _start_data_json(self) -> str:
        """
        Output json with start data to write for external revocation registry builder process pickup.

        :return: logging and wallet init data json
        """

        rv = {
            'logging': {
                'paths': []
            },
            'wallet': {
            }
        }

        logger = LOGGER
        while not logger.level:
            logger = logger.parent
            if logger is None:
                break
        rv['logging']['level'] = logger.level

        logger = LOGGER
        log_paths = [realpath(h.baseFilename) for h in logger.handlers if hasattr(h, 'baseFilename')]
        while not log_paths:
            logger = logger.parent
            if logger is None:
                break
            log_paths = [realpath(h.baseFilename) for h in logger.handlers if hasattr(h, 'baseFilename')]
        for log_path in log_paths:
            rv['logging']['paths'].append(log_path)

        rv['wallet']['storage_type'] = self.wallet.storage_type
        rv['wallet']['config'] = self.wallet.config
        rv['wallet']['access_creds'] = self.wallet.access_creds

        return json.dumps(rv)

    @staticmethod
    def _get_state(wallet_name: str) -> _STATE:
        """
        Return current state of revocation registry builder process.

        :param wallet_name: name of wallet for corresponding Issuer
        :return: current process state as _STATE enum
        """

        dir_sentinel = RevRegBuilder.dir_tails_sentinel(wallet_name)
        file_pid = join(dir_sentinel, '.pid')
        file_start = join(dir_sentinel, '.start')
        file_stop = join(dir_sentinel, '.stop')

        if isfile(file_stop):
            return _STATE.STOPPING

        if isfile(file_start) or isfile(file_pid):
            return _STATE.RUNNING

        return _STATE.ABSENT

    @staticmethod
    def dir_tails_sentinel(wallet_name: str) -> str:
        """
        Return the sentinel directory for a revocation registry builder on input wallet name, without
        instantiating the anchor.

        :param wallet_name: name of revocation registry builder, as specified in its wallet configuration
        :return: path to sentinel directory for revocation registry builder on wallet name
        """

        return join(RevRegBuilder.DIR_TAILS, '.sentinel', wallet_name)

    def dir_tails_top(self, rr_id) -> str:
        """
        Return top of tails tree for input rev reg id.

        :param rr_id: revocation registry identifier
        :return: top of tails tree
        """

        return join(self.dir_tails_hopper, rr_id) if self.external else self.dir_tails

    def dir_tails_target(self, rr_id) -> str:
        """
        Return target directory for revocation registry and tails file production.

        :param rr_id: revocation registry identifier
        :return: tails target directory
        """

        return join(self.dir_tails_top(rr_id), rev_reg_id2cred_def_id(rr_id))

    def mark_in_progress(self, rr_id: str, rr_size: int) -> None:
        """
        Prepare sentinel directory for revocation registry construction.

        :param rr_id: revocation registry identifier
        :rr_size: size of revocation registry to build
        """
        try:
            makedirs(join(self._dir_tails_sentinel, rr_id), exist_ok=False)
        except FileExistsError:
            LOGGER.warning('Rev reg %s construction already in progress', rr_id)
        else:
            open(join(self._dir_tails_sentinel, rr_id, '.{}'.format(rr_size)), 'w').close()

    async def serve(self) -> None:
        """
        Write pidfile to sentinel directory if need be, and wait for sentinels
        to shut down or build revocation registry and tails file.
        """

        LOGGER.debug('RevRegBuilder.serve >>>')

        assert self.external

        file_pid = join(self._dir_tails_sentinel, '.pid')
        if isfile(file_pid):
            with open(file_pid, 'r') as fh_pid:
                pid = int(fh_pid.read())
            try:
                kill(pid, 0)
            except ProcessLookupError:
                remove(file_pid)
                LOGGER.info('RevRegBuilder removed derelict .pid file')
            except PermissionError:
                LOGGER.info('RevRegBuilder process already running with pid %s: exiting', pid)
                LOGGER.debug('RevRegBuilder.serve <<<')
                return
            else:
                LOGGER.info('RevRegBuilder process already running with pid %s: exiting', pid)
                LOGGER.debug('RevRegBuilder.serve <<<')
                return

        pid = getpid()
        with open(file_pid, 'w') as pid_fh:
            print(str(pid), file=pid_fh)

        file_stop = join(self._dir_tails_sentinel, '.stop')

        while True:
            if isfile(file_stop):  # stop now, pick up any pending tasks next invocation
                remove(file_stop)
                remove(file_pid)
                break

            p_pending = [join(self._dir_tails_sentinel, d) for d in listdir(self._dir_tails_sentinel)
                if isdir(join(self._dir_tails_sentinel, d))]
            p_pending = [p for p in p_pending if [s for s in listdir(p) if s.startswith('.')]]  # size marker
            if p_pending:
                pdir = basename(p_pending[0])
                rr_id = pdir
                rr_size = int([s for s in listdir(p_pending[0]) if s.startswith('.')][0][1:])
                open(join(p_pending[0], '.in-progress'), 'w').close()
                await self.create_rev_reg(rr_id, rr_size or None)
                rmtree(p_pending[0])
            await asyncio.sleep(1)

        LOGGER.debug('RevRegBuilder.serve <<<')

    @staticmethod
    async def stop(wallet_name: str) -> None:
        """
        Gracefully stop an external revocation registry builder, waiting for its current.

        The indy-sdk toolkit uses a temporary directory for tails file mustration,
        and shutting down the toolkit removes the directory, crashing the external
        tails file write. This method allows a graceful stop to wait for completion
        of such tasks already in progress.

        :wallet_name: name external revocation registry builder to check
        :return: whether a task is pending.
        """

        LOGGER.debug('RevRegBuilder.stop >>>')

        dir_sentinel = join(RevRegBuilder.dir_tails_sentinel(wallet_name))

        if isdir(dir_sentinel):
            open(join(dir_sentinel, '.stop'), 'w').close()  # touch

            while any(isfile(join(dir_sentinel, d, '.in-progress')) for d in listdir(dir_sentinel)):
                await asyncio.sleep(1)

        LOGGER.debug('RevRegBuilder.stop <<<')

    async def create_rev_reg(self, rr_id: str, rr_size: int = None) -> None:
        """
        Create revocation registry artifacts and new tails file (with association to
        corresponding revocation registry identifier via symbolic link name)
        for input revocation registry identifier. Symbolic link presence signals completion.
        If revocation registry builder operates in a process external to its Issuer's,
        target directory is hopper directory.

        Raise WalletState for closed wallet.

        :param rr_id: revocation registry identifier
        :param rr_size: revocation registry size (defaults to 64)
        """

        LOGGER.debug('RevRegBuilder.create_rev_reg >>> rr_id: %s, rr_size: %s', rr_id, rr_size)

        if not self.wallet.handle:
            LOGGER.debug('RevRegBuilder.create_rev_reg <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('RevRegBuilder.create_rev_reg <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        rr_size = rr_size or 64

        (cd_id, tag) = rev_reg_id2cred_def_id_tag(rr_id)

        dir_tails = self.dir_tails_top(rr_id)
        dir_target = self.dir_tails_target(rr_id)
        if self.external:
            try:
                makedirs(dir_target, exist_ok=False)
            except FileExistsError:
                LOGGER.warning(
                    'RevRegBuilder.create_rev_reg found dir %s, but task not in progress: rebuilding rev reg %s',
                    dir_target,
                    rr_id)
                rmtree(dir_target)
                makedirs(dir_target, exist_ok=False)

        LOGGER.info('Creating revocation registry (capacity %s) for rev reg id %s', rr_size, rr_id)
        tails_writer_handle = await blob_storage.open_writer(
            'default',
            json.dumps({
                'base_dir': dir_target,
                'uri_pattern': ''
            }))

        (created_rr_id, rr_def_json, rr_ent_json) = await anoncreds.issuer_create_and_store_revoc_reg(
            self.wallet.handle,
            self.did,
            'CL_ACCUM',
            tag,
            cd_id,
            json.dumps({
                'max_cred_num': rr_size,
                'issuance_type': 'ISSUANCE_BY_DEFAULT'
            }),
            tails_writer_handle)

        tails_hash = basename(Tails.unlinked(dir_target).pop())
        with open(join(dir_target, 'rr_def.json'), 'w') as rr_def_fh:
            print(rr_def_json, file=rr_def_fh)
        with open(join(dir_target, 'rr_ent.json'), 'w') as rr_ent_fh:
            print(rr_ent_json, file=rr_ent_fh)
        Tails.associate(dir_tails, created_rr_id, tails_hash)  # associate last: symlink signals completion

        LOGGER.debug('RevRegBuilder.create_rev_reg <<<')


async def main(wallet_name: str) -> None:
    """
    Main line for revocation registry builder operating in external process on behalf of issuer agent.

    :param wallet_name: wallet name - must match that of issuer with existing wallet
    """

    logging.basicConfig(level=logging.WARN, format='%(levelname)-8s | %(name)-12s | %(message)s')
    logging.getLogger('indy').setLevel(logging.ERROR)

    path_start = join(RevRegBuilder.dir_tails_sentinel(wallet_name), '.start')

    with open(path_start, 'r') as fh_start:
        start_data = json.loads(fh_start.read())
    remove(path_start)

    logging.getLogger(__name__).setLevel(start_data['logging']['level'])
    for path_log in start_data['logging']['paths']:
        logging.getLogger(__name__).addHandler(logging.FileHandler(path_log))

    wallet = WalletManager().get(
        {
            'id': wallet_name,
            'storage_type': start_data['wallet']['storage_type'],
            **start_data['wallet']['config'],
        },
        access=start_data['wallet']['access_creds'].get('key', None))
    async with wallet, RevRegBuilder(wallet, rrbx=True) as rrban:
        await rrban.serve()


if __name__ == '__main__':
    PARSER = argparse.ArgumentParser()
    PARSER.add_argument('-n', '--name', help='wallet name', required=True)
    ARGS = PARSER.parse_args()

    LOOP = asyncio.get_event_loop()
    LOOP.run_until_complete(main(ARGS.name))



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from time import time
from typing import Callable, Sequence, Union

from indy import anoncreds, crypto, did, non_secrets, wallet
from indy.error import IndyError, ErrorCode

from von_anchor.canon import canon_non_secret_wql, canon_pairwise_wql
from von_anchor.error import (
    AbsentRecord,
    AbsentMessage,
    AbsentWallet,
    BadAccess,
    BadKey,
    BadIdentifier,
    BadRecord,
    ExtantRecord,
    ExtantWallet,
    WalletState)
from von_anchor.util import ok_did
from von_anchor.wallet import DIDInfo, KeyInfo, storage_record2pairwise_info, PairwiseInfo, pairwise_info2tags
from von_anchor.wallet.record import StorageRecord, TYPE_PAIRWISE, TYPE_LINK_SECRET_LABEL


LOGGER = logging.getLogger(__name__)


class Wallet:
    """
    Class encapsulating indy-sdk wallet.
    """

    DEFAULT_CHUNK = 256  # chunk size in searching credentials, non-secret storage records

    def __init__(self, indy_config: dict, von_config: dict) -> None:
        """
        Initializer for wallet that WalletManager created. Store configuration and access credentials value.

        Actuators should prefer WalletManager.get() to calling this initializer directly - the wallet manager
        filters wallet configuration through preset defaults.

        :param indy_config: configuration for indy-sdk wallet
        :param von_config: VON wallet configuration particulars:

            - 'seed': (optional) seed to use on creation
            - 'did': (optional) anchor DID to use on creation
            - 'link_secret_label': (optional) label to use to create link secret
            - 'auto_create': whether to create wallet automatically on first open
            - 'auto_remove': whether to remove wallet automatically on next close
            - 'access': wallet access credentials value

        """

        LOGGER.debug('Wallet.__init__ >>> indy_config %s, von_config %s', indy_config, von_config)

        self._handle = None
        self._indy_config = {**indy_config}  # make a copy
        self._von_config = {**von_config}
        self._did = None
        self._verkey = None

        LOGGER.debug('Wallet.__init__ <<<')

    @property
    def name(self) -> str:
        """
        Accessor for wallet name, as configuration retains at key 'id'.

        :return: wallet name
        """

        return self.config['id']

    @property
    def handle(self) -> int:
        """
        Accessor for indy-sdk wallet handle.

        :return: indy-sdk wallet handle
        """

        return self._handle

    @property
    def opened(self) -> bool:
        """
        Accessor for indy-sdk wallet state: True for open, False for closed.

        :return: indy-sdk wallet state
        """

        return bool(self._handle)

    @property
    def config(self) -> dict:
        """
        Accessor for wallet config.

        :return: wallet config
        """

        return self._indy_config

    @property
    def auto_create(self) -> bool:
        """
        Accessor for auto_create wallet config setting.

        :return: auto_create wallet config setting
        """

        return self._von_config['auto_create']

    @auto_create.setter
    def auto_create(self, value: bool) -> None:
        """
        Set auto_create wallet config behaviour.

        :param value: auto_create
        """

        self._von_config['auto_create'] = value

    @property
    def auto_remove(self) -> bool:
        """
        Accessor for auto_remove wallet config setting.

        :return: auto_remove wallet config setting
        """

        return self._von_config['auto_remove']

    @auto_remove.setter
    def auto_remove(self, value: bool) -> None:
        """
        Set auto_remove wallet config behaviour.

        :param value: auto_remove
        """

        self._von_config['auto_remove'] = value

    @property
    def access_creds(self) -> dict:
        """
        Accessor for wallet access credentials.

        :return: wallet access credentials
        """

        return {'key': self._von_config['access']}

    @property
    def access(self) -> str:
        """
        Accessor for wallet access credentials value.

        :return: wallet access credentials value
        """

        return self._von_config['access']

    @property
    def storage_type(self) -> str:
        """
        Accessor for wallet type, as configuration retains at key 'storage_type'.

        :return: wallet type
        """

        return self.config['storage_type']

    @property
    def did(self) -> str:
        """
        Accessor for anchor DID in wallet.

        :return: anchor DID in wallet
        """

        return self._did

    @did.setter
    def did(self, value: str) -> None:
        """
        Set anchor DID in wallet.

        :param value: anchor DID
        """

        self._did = value

    @property
    def verkey(self) -> str:
        """
        Accessor for wallet verification key.

        :return: wallet verification key
        """

        return self._verkey

    @verkey.setter
    def verkey(self, value: str) -> None:
        """
        Set verification key.

        :param value: verification key
        """

        self._verkey = value

    async def create_signing_key(self, seed: str = None, metadata: dict = None) -> KeyInfo:
        """
        Create a new signing key pair.

        Raise WalletState if wallet is closed, ExtantRecord if verification key already exists.

        :param seed: optional seed allowing deterministic key creation
        :param metadata: optional metadata to store with key pair
        :return: KeyInfo for new key pair
        """

        LOGGER.debug('Wallet.create_signing_key >>> seed: [SEED], metadata: %s', metadata)

        if not self.handle:
            LOGGER.debug('Wallet.create_signing_key <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            verkey = await crypto.create_key(self.handle, json.dumps({'seed': seed} if seed else {}))
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemAlreadyExists:
                LOGGER.debug('Wallet.create_signing_key <!< Verification key already present in wallet %s', self.name)
                raise ExtantRecord('Verification key already present in wallet {}'.format(self.name))
            LOGGER.debug('Wallet.create_signing_key <!< indy-sdk raised error %s', x_indy.error_code)
            raise

        await crypto.set_key_metadata(self.handle, verkey, json.dumps(metadata or {}))  # coerce None to empty

        rv = KeyInfo(verkey, metadata or {})
        LOGGER.debug('Wallet.create_signing_key <<< %s', rv)
        return rv

    async def get_signing_key(self, verkey: str) -> KeyInfo:
        """
        Get signing key pair for input verification key.

        Raise WalletState if wallet is closed, AbsentRecord for no such key pair.

        :param verkey: verification key of key pair
        :return: KeyInfo for key pair
        """

        LOGGER.debug('Wallet.get_signing_key >>> seed: [SEED], verkey: %s', verkey)

        if not self.handle:
            LOGGER.debug('Wallet.get_signing_key <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            metadata = await crypto.get_key_metadata(self.handle, verkey)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                LOGGER.debug('Wallet.get_signing_key <!< Verification key %s not in wallet %s', verkey, self.name)
                raise AbsentRecord('Verification key not in wallet {}'.format(self.name))
            LOGGER.debug('Wallet.get_signing_key <!< indy-sdk raised error %s', x_indy.error_code)
            raise

        rv = KeyInfo(verkey, json.loads(metadata) if metadata else {})
        LOGGER.debug('Wallet.get_signing_key <<< %s', rv)
        return rv

    async def replace_signing_key_metadata(self, verkey: str, metadata: dict) -> KeyInfo:
        """
        Replace the metadata associated with a signing key pair.

        Raise WalletState if wallet is closed, AbsentRecord for no such key pair.

        :param verkey: verification key of key pair
        :param metadata: new metadata to store
        :return: resulting KeyInfo for key pair
        """

        LOGGER.debug('Wallet.replace_signing_key_metadata >>> verkey: %s, metadata: %s', verkey, metadata)

        if not self.handle:
            LOGGER.debug('Wallet.replace_signing_key_metadata <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            await crypto.get_key_metadata(self.handle, verkey)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                LOGGER.debug(
                    'Wallet.replace_signing_key_metadata <!< Verification key %s not in wallet %s',
                    verkey,
                    self.name)
                raise AbsentRecord('Verification key not in wallet {}'.format(self.name))
            LOGGER.debug('Wallet.replace_signing_key_metadata <!< indy-sdk raised error %s', x_indy.error_code)
            raise

        await crypto.set_key_metadata(self.handle, verkey, json.dumps(metadata or {}))

        rv = await self.get_signing_key(verkey)
        LOGGER.debug('Wallet.replace_signing_key_metadata <<< %s', rv)
        return rv

    async def create_local_did(self, seed: str = None, loc_did: str = None, metadata: dict = None) -> DIDInfo:
        """
        Create and store a new local DID for use in pairwise DID relations.

        :param seed: seed from which to create (default random)
        :param loc_did: local DID value (default None to let indy-sdk generate)
        :param metadata: metadata to associate with the local DID
            (operation always sets 'since', 'modified' epoch timestamps)
        :return: DIDInfo for new local DID
        """

        LOGGER.debug('Wallet.create_local_did >>> seed: [SEED] loc_did: %s metadata: %s', loc_did, metadata)

        cfg = {}
        if seed:
            cfg['seed'] = seed
        if loc_did:
            cfg['did'] = loc_did

        if not self.handle:
            LOGGER.debug('Wallet.create_local_did <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            (created_did, verkey) = await did.create_and_store_my_did(self.handle, json.dumps(cfg))
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.DidAlreadyExistsError:
                LOGGER.debug('Wallet.create_local_did <!< DID %s already present in wallet %s', loc_did, self.name)
                raise ExtantRecord('Local DID {} already present in wallet {}'.format(loc_did, self.name))
            LOGGER.debug('Wallet.create_local_did <!< indy-sdk raised error %s', x_indy.error_code)
            raise

        now = int(time())
        loc_did_metadata = {**(metadata or {}), 'since': now, 'modified': now}
        await did.set_did_metadata(self.handle, created_did, json.dumps(loc_did_metadata))

        rv = DIDInfo(created_did, verkey, loc_did_metadata)

        LOGGER.debug('Wallet.create_local_did <<< %s', rv)
        return rv

    async def replace_local_did_metadata(self, loc_did: str, metadata: dict) -> DIDInfo:
        """
        Replace the metadata associated with a local DID.

        Raise WalletState if wallet is closed, AbsentRecord for no such local DID.

        :param loc_did: local DID of interest
        :param metadata: new metadata to store
        :return: DIDInfo for local DID after write
        """

        LOGGER.debug('Wallet.replace_local_did_metadata >>> loc_did: %s, metadata: %s', loc_did, metadata)

        old = await self.get_local_did(loc_did)  # raises exceptions if applicable
        now = int(time())
        loc_did_metadata = {**(metadata or {}), 'since': (old.metadata or {}).get('since', now), 'modified': now}
        try:
            await did.set_did_metadata(self.handle, loc_did, json.dumps(loc_did_metadata))
        except IndyError as x_indy:
            LOGGER.debug('Wallet.replace_local_did_metadata <!< indy-sdk raised error %s', x_indy.error_code)
            raise

        rv = await self.get_local_did(loc_did)
        LOGGER.debug('Wallet.replace_local_did_metadata <<< %s', rv)
        return rv

    async def get_local_dids(self) -> Sequence[DIDInfo]:
        """
        Get list of DIDInfos for local DIDs.

        :return: list of local DIDInfos
        """

        LOGGER.debug('Wallet.get_local_dids >>>')

        dids_with_meta = json.loads(did.list_my_dids_with_meta(self.handle))  # list

        rv = []
        for did_with_meta in dids_with_meta:
            meta = json.loads(did_with_meta['metadata']) if did_with_meta['metadata'] else {}
            if meta.get('anchor', False):
                continue  # exclude anchor DIDs past and present
            rv.append(DIDInfo(did_with_meta['did'], did_with_meta['verkey'], meta))

        LOGGER.debug('Wallet.get_local_dids <<< %s', rv)
        return rv

    async def get_local_did(self, loc: str) -> DIDInfo:
        """
        Get local DID info by local DID or verification key.
        Raise AbsentRecord for no such local DID.

        :param loc: DID or verification key of interest
        :return: DIDInfo for local DID
        """

        LOGGER.debug('Wallet.get_local_did >>> loc: %s', loc)

        if not self.handle:
            LOGGER.debug('Wallet.get_local_did <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        if ok_did(loc):  # it's a DID
            try:
                did_with_meta = json.loads(await did.get_my_did_with_meta(self.handle, loc))
                rv = DIDInfo(
                    did_with_meta['did'],
                    did_with_meta['verkey'],
                    json.loads(did_with_meta['metadata']) if did_with_meta['metadata'] else {})  # nudge None to empty
            except IndyError as x_indy:
                if x_indy.error_code == ErrorCode.WalletItemNotFound:
                    LOGGER.debug('Wallet.get_local_did <!< DID %s not present in wallet %s', loc, self.name)
                    raise AbsentRecord('Local DID {} not present in wallet {}'.format(loc, self.name))
                LOGGER.debug('Wallet.get_local_did <!< indy-sdk raised error %s', x_indy.error_code)
                raise

        else:  # it's a verkey
            dids_with_meta = json.loads(await did.list_my_dids_with_meta(self.handle))  # list
            for did_with_meta in dids_with_meta:
                if did_with_meta['verkey'] == loc:
                    rv = DIDInfo(
                        did_with_meta['did'],
                        did_with_meta['verkey'],
                        json.loads(did_with_meta['metadata']) if did_with_meta['metadata'] else {})
                    break
            else:
                LOGGER.debug('Wallet.get_local_did <!< Wallet %s has no local DID for verkey %s', self.name, loc)
                raise AbsentRecord('Wallet {} has no local DID for verkey {}'.format(self.name, loc))

        LOGGER.debug('Wallet.get_local_did <<< %s', rv)
        return rv

    async def get_anchor_did(self) -> str:
        """
        Get current anchor DID by metadata, None for not yet set.

        :return: DID
        """

        LOGGER.debug('Wallet.get_anchor_did >>>')

        if not self.handle:
            LOGGER.debug('Wallet.get_anchor_did <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = None
        dids_with_meta = json.loads(await did.list_my_dids_with_meta(self.handle))  # list

        latest = 0
        for did_with_meta in dids_with_meta:
            try:
                meta = json.loads(did_with_meta['metadata']) if did_with_meta['metadata'] else {}
                if not meta.get('anchor', False):
                    continue
                if isinstance(meta, dict) and meta.get('since', -1) > latest:
                    rv = did_with_meta.get('did')
            except json.decoder.JSONDecodeError:
                continue  # it's not an anchor DID, carry on

        LOGGER.debug('Wallet.get_anchor_did <<< %s', rv)
        return rv

    async def create_link_secret(self, label: str) -> None:
        """
        Create link secret (a.k.a. master secret) used in proofs by HolderProver, if the
        current link secret does not already correspond to the input link secret label.

        Raise WalletState if wallet is closed, or any other IndyError causing failure
        to set link secret in wallet.

        :param label: label for link secret; indy-sdk uses label to generate link secret
        """

        LOGGER.debug('Wallet.create_link_secret >>> label: %s', label)

        if not self.handle:
            LOGGER.debug('Wallet.create_link_secret <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            await anoncreds.prover_create_master_secret(self.handle, label)
            await self._write_link_secret_label(label)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.AnoncredsMasterSecretDuplicateNameError:
                LOGGER.warning(
                    'Wallet %s link secret already current: abstaining from updating label record', self.name)
                await self._write_link_secret_label(label)
            else:
                LOGGER.debug(
                    'Wallet.create_link_secret <!< cannot create link secret for wallet %s, indy error code %s',
                    self.name,
                    x_indy.error_code)
                raise

        LOGGER.debug('Wallet.create_link_secret <<<')

    async def _write_link_secret_label(self, label) -> None:
        """
        Update non-secret storage record with link secret label.

        :param label: link secret label
        """

        LOGGER.debug('Wallet._write_link_secret_label <<< %s', label)

        if await self.get_link_secret_label() == label:
            LOGGER.info('Wallet._write_link_secret_label abstaining - already current')
        else:
            await self.write_non_secret(StorageRecord(
                TYPE_LINK_SECRET_LABEL,
                label,
                tags=None,
                ident=str(int(time()))))  # indy requires str

        LOGGER.debug('Wallet._write_link_secret_label <<<')

    async def get_link_secret_label(self) -> str:
        """
        Get current link secret label from non-secret storage records; return None for no match.

        :return: latest non-secret storage record for link secret label
        """

        LOGGER.debug('Wallet.get_link_secret_label >>>')

        if not self.handle:
            LOGGER.debug('Wallet.get_link_secret <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = None
        records = await self.get_non_secret(TYPE_LINK_SECRET_LABEL)
        if records:
            rv = records[str(max(int(k) for k in records))].value  # str to int, max, and back again

        LOGGER.debug('Wallet.get_link_secret_label <<< %s', rv)
        return rv

    async def __aenter__(self) -> 'Wallet':
        """
        Context manager entry. Open wallet as configured, for closure on context manager exit.
        For use in monolithic call opening, using, and closing wallet.

        Raise any IndyError causing failure to open wallet, or AbsentWallet on attempt to enter wallet
        not yet created.

        :return: current object
        """

        LOGGER.debug('Wallet.__aenter__ >>>')

        rv = await self.open()
        LOGGER.debug('Wallet.__aenter__ <<<')
        return rv

    async def open(self) -> 'Wallet':
        """
        Explicit entry. Open wallet as configured, for later closure via close().
        For use when keeping wallet open across multiple calls.

        Raise any IndyError causing failure to open wallet, WalletState if wallet already open,
        or AbsentWallet on attempt to enter wallet not yet created.

        :return: current object
        """

        LOGGER.debug('Wallet.open >>>')

        created = False
        while True:
            try:
                self._handle = await wallet.open_wallet(
                    json.dumps(self.config),
                    json.dumps(self.access_creds))
                LOGGER.info('Opened wallet %s on handle %s', self.name, self.handle)
                break
            except IndyError as x_indy:
                if x_indy.error_code == ErrorCode.WalletNotFoundError:
                    if created:
                        LOGGER.debug('Wallet.open() <!< Wallet %s not found after creation', self.name)
                        raise AbsentWallet('Wallet {} not found after creation'.format(self.name))
                    if self.auto_create:
                        await self.create()
                        continue
                    else:
                        LOGGER.debug('Wallet.open() <!< Wallet %s not found', self.name)
                        raise AbsentWallet('Wallet {} not found'.format(self.name))
                elif x_indy.error_code == ErrorCode.WalletAlreadyOpenedError:
                    LOGGER.debug('Wallet.open() <!< Wallet %s is already open', self.name)
                    raise WalletState('Wallet {} is already open'.format(self.name))
                elif x_indy.error_code == ErrorCode.WalletAccessFailed:
                    LOGGER.debug('Wallet.open() <!< Bad access credentials value for wallet %s', self.name)
                    raise BadAccess('Bad access credentials value for wallet {}'.format(self.name))

                LOGGER.debug('Wallet %s open raised indy error %s', self.name, x_indy.error_code)
                raise

        self.did = await self.get_anchor_did()
        self.verkey = await did.key_for_local_did(self.handle, self.did) if self.did else None
        LOGGER.info('Wallet %s got verkey %s for existing DID %s', self.name, self.verkey, self.did)

        LOGGER.debug('Wallet.open <<<')
        return self

    async def create(self) -> None:
        """
        Persist the wallet. Raise ExtantWallet if it already exists.

        Actuators should prefer WalletManager.create() to calling this method directly - the wallet manager
        filters wallet configuration through preset defaults.
        """

        LOGGER.debug('Wallet.create >>>')

        try:
            await wallet.create_wallet(
                config=json.dumps(self.config),
                credentials=json.dumps(self.access_creds))
            LOGGER.info('Created wallet %s', self.name)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletAlreadyExistsError:
                LOGGER.debug('Wallet.create <!< Wallet %s already exists', self.name)
                raise ExtantWallet('Wallet {} already exists'.format(self.name))
            LOGGER.debug(
                'Wallet.create <!< indy error code %s on creation of wallet %s',
                x_indy.error_code,
                self.name)
            raise

        auto_remove = self.auto_remove
        self.auto_remove = False  # defer past this creation process
        async with self:
            did_info = await self.create_local_did(
                self._von_config.get('seed', None),
                self._von_config.get('did', None),
                {'anchor': True})
            self.did = did_info.did
            self.verkey = did_info.verkey
            if 'link_secret_label' in self._von_config:
                await self.create_link_secret(self._von_config['link_secret_label'])
        self.auto_remove = auto_remove

        LOGGER.debug('Wallet.create <<<')

    async def __aexit__(self, exc_type, exc, traceback) -> None:
        """
        Context manager exit. Close wallet (and delete if so configured).
        For use in monolithic call opening, using, and closing the wallet.

        :param exc_type:
        :param exc:
        :param traceback:
        """

        LOGGER.debug('Wallet.__aexit__ >>>')

        await self.close()

        LOGGER.debug('Wallet.__aexit__ <<<')

    async def close(self) -> None:
        """
        Explicit exit. Close wallet (and delete if so configured).
        """

        LOGGER.debug('Wallet.close >>>')

        if not self.handle:
            LOGGER.warning('Abstaining from closing wallet %s: already closed', self.name)
        else:
            LOGGER.debug('Closing wallet %s', self.name)
            await wallet.close_wallet(self.handle)
            self._handle = None
            if self.auto_remove:
                LOGGER.info('Automatically removing wallet %s', self.name)
                await self.remove()
        self._handle = None

        LOGGER.debug('Wallet.close <<<')

    async def remove(self) -> bool:
        """
        Remove serialized wallet, best effort, if it exists. Return whether wallet absent after operation
        (removal successful or else not present a priori).

        Raise WalletState if wallet is open.

        :return: whether wallet gone from persistent storage
        """

        LOGGER.debug('Wallet.remove >>>')

        if self.handle:
            LOGGER.debug('Wallet.remove <!< Wallet %s is open', self.name)
            raise WalletState('Wallet {} is open'.format(self.name))

        rv = True
        try:
            LOGGER.info('Attempting to remove wallet: %s', self.name)
            await wallet.delete_wallet(
                json.dumps(self.config),
                json.dumps(self.access_creds))
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletNotFoundError:
                LOGGER.info('Wallet %s not present; abstaining from removal', self.name)
            else:
                LOGGER.info('Failed wallet %s removal; indy-sdk error code %s', self.name, x_indy.error_code)
                rv = False

        LOGGER.debug('Wallet.remove <<< %s', rv)
        return rv

    async def write_pairwise(
            self,
            their_did: str,
            their_verkey: str = None,
            my_did: str = None,
            metadata: dict = None,
            replace_meta: bool = False) -> PairwiseInfo:
        """
        Store a pairwise DID for a secure connection. Use verification key for local DID in wallet if
        supplied; otherwise, create one first. If local DID specified but not present, raise AbsentRecord.

        With supplied metadata, replace or augment and overwrite any existing metadata for the pairwise
        relation if one already exists in the wallet. Always include local and remote DIDs and keys in
        metadata to allow for WQL search.

        Raise AbsentRecord on call to update a non-existent record. Raise BadRecord if metadata does not
        coerce into non-secrets API tags specification {str:str}.

        :param their_did: remote DID
        :param their_verkey: remote verification key (default None is OK if updating an existing pairwise DID)
        :param my_did: local DID (default None prompts operation to generate one at random)
        :param metadata: metadata for pairwise connection
        :param replace_meta: whether to (True) replace or (False) augment and overwrite existing metadata
        :return: resulting PairwiseInfo
        """

        LOGGER.debug(
            'Wallet.write_pairwise >>> their_did: %s, their_verkey: %s, my_did: %s, metadata: %s, replace_meta: %s',
            their_did,
            their_verkey,
            my_did,
            metadata,
            replace_meta)

        if their_verkey is None:
            match = await self.get_pairwise(their_did)
            if not match:
                LOGGER.debug(
                    'Wallet.write_pairwise <!< Wallet %s has no pairwise DID on %s to update',
                    self.name,
                    their_did)
                raise AbsentRecord('Wallet {} has no pairwise DID on {} to update'.format(self.name, their_did))
            their_verkey = [pwise for pwise in match.values()][0].their_verkey

        try:
            await did.store_their_did(self.handle, json.dumps({'did': their_did, 'verkey': their_verkey}))
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemAlreadyExists:
                pass  # exists already, carry on
            else:
                LOGGER.debug(
                    'Wallet.write_pairwise <!< Wallet %s write of their_did %s raised indy error code %s',
                    self.name,
                    their_did,
                    x_indy.error_code)
                raise

        if my_did:
            my_did_info = await self.get_local_did(my_did)  # raises AbsentRecord if no such local did
        else:
            my_did_info = await self.create_local_did(None, None, {'pairwise_for': their_did})

        pairwise = PairwiseInfo(their_did, their_verkey, my_did_info.did, my_did_info.verkey, metadata)
        try:
            storec = await self.write_non_secret(
                StorageRecord(TYPE_PAIRWISE, their_verkey, tags=pairwise_info2tags(pairwise), ident=their_did),
                replace_meta)
        except BadRecord:
            LOGGER.debug(
                'Wallet.write_pairwise <!< Pairwise metadata %s does not coerce into flat {str:str} tags dict',
                pairwise.metadata)
            raise

        rv = storage_record2pairwise_info(storec)
        LOGGER.debug('Wallet.write_pairwise <<< %s', rv)
        return rv

    async def delete_pairwise(self, their_did: str) -> None:
        """
        Remove a pairwise DID record by its remote DID. Silently return if no such record is present.
        Raise WalletState for closed wallet, or BadIdentifier for invalid pairwise DID.

        :param their_did: remote DID marking pairwise DID to remove
        """

        LOGGER.debug('Wallet.delete_pairwise >>> their_did: %s', their_did)

        if not ok_did(their_did):
            LOGGER.debug('Wallet.delete_pairwise <!< Bad DID %s', their_did)
            raise BadIdentifier('Bad DID {}'.format(their_did))

        await self.delete_non_secret(TYPE_PAIRWISE, their_did)

        LOGGER.debug('Wallet.delete_pairwise <<<')

    async def get_pairwise(self, pairwise_filt: str = None) -> dict:
        """
        Return dict mapping each pairwise DID of interest in wallet to its pairwise info, or,
        for no filter specified, mapping them all. If wallet has no such item, return empty dict.

        :param pairwise_filt: remote DID of interest, or WQL json (default all)
        :return: dict mapping remote DIDs to PairwiseInfo
        """

        LOGGER.debug('Wallet.get_pairwise >>> pairwise_filt: %s', pairwise_filt)

        if not self.handle:
            LOGGER.debug('Wallet.get_pairwise <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        storecs = await self.get_non_secret(
            TYPE_PAIRWISE,
            pairwise_filt if not pairwise_filt or ok_did(pairwise_filt) else json.loads(pairwise_filt),
            canon_pairwise_wql)
        rv = {k: storage_record2pairwise_info(storecs[k]) for k in storecs}  # touch up tags, mute leading ~

        LOGGER.debug('Wallet.get_pairwise <<< %s', rv)
        return rv

    async def write_non_secret(self, storec: StorageRecord, replace_meta: bool = False) -> StorageRecord:
        """
        Add or update non-secret storage record to the wallet; return resulting wallet non-secret record.

        :param storec: non-secret storage record
        :param replace_meta: whether to replace any existing metadata on matching record or to augment it
        :return: non-secret storage record as it appears in the wallet after write
        """

        LOGGER.debug('Wallet.write_non_secret >>> storec: %s, replace_meta: %s', storec, replace_meta)

        if not self.handle:
            LOGGER.debug('Wallet.write_non_secret <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        if not StorageRecord.ok_tags(storec.tags):
            LOGGER.debug('Wallet.write_non_secret <!< bad storage record tags %s; use flat {str: str} dict', storec)
            raise BadRecord('Bad storage record tags {}; use flat {{str:str}} dict'.format(storec))

        try:
            record = json.loads(await non_secrets.get_wallet_record(
                self.handle,
                storec.type,
                storec.id,
                json.dumps({
                    'retrieveType': False,
                    'retrieveValue': True,
                    'retrieveTags': True
                })))
            if record['value'] != storec.value:
                await non_secrets.update_wallet_record_value(
                    self.handle,
                    storec.type,
                    storec.id,
                    storec.value)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                await non_secrets.add_wallet_record(
                    self.handle,
                    storec.type,
                    storec.id,
                    storec.value,
                    json.dumps(storec.tags) if storec.tags else None)
            else:
                LOGGER.debug(
                    'Wallet.write_non_secret <!< Wallet lookup raised indy error code %s',
                    x_indy.error_code)
                raise
        else:
            if (record['tags'] or None) != storec.tags:  # record maps no tags to {}, not None
                tags = (storec.tags or {}) if replace_meta else {**record['tags'], **(storec.tags or {})}

                await non_secrets.update_wallet_record_tags(
                    self.handle,
                    storec.type,
                    storec.id,
                    json.dumps(tags))  # indy-sdk takes '{}' instead of None for null tags

        record = json.loads(await non_secrets.get_wallet_record(
            self.handle,
            storec.type,
            storec.id,
            json.dumps({
                'retrieveType': False,
                'retrieveValue': True,
                'retrieveTags': True
            })))

        rv = StorageRecord(storec.type, record['value'], tags=record.get('tags', None), ident=record['id'])
        LOGGER.debug('Wallet.write_non_secret <<< %s', rv)
        return rv

    async def delete_non_secret(self, typ: str, ident: str) -> None:
        """
        Remove a non-secret record by its type and identifier. Silently return if no such record is present.
        Raise WalletState for closed wallet.

        :param typ: non-secret storage record type
        :param ident: non-secret storage record identifier
        """

        LOGGER.debug('Wallet.delete_non_secret >>> typ: %s, ident: %s', typ, ident)

        if not self.handle:
            LOGGER.debug('Wallet.delete_non_secret <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        try:
            await non_secrets.delete_wallet_record(self.handle, typ, ident)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                LOGGER.info('Wallet.delete_non_secret <!< no record for type %s on identifier %s', typ, ident)
            else:
                LOGGER.debug(
                    'Wallet.delete_non_secret <!< deletion of %s record on identifier %s raised indy error code %s',
                    typ,
                    ident,
                    x_indy.error_code)
                raise

        LOGGER.debug('Wallet.delete_non_secret <<<')

    async def get_non_secret(
            self,
            typ: str,
            filt: Union[dict, str] = None,
            canon_wql: Callable[[dict], dict] = None,
            limit: int = None) -> dict:
        """
        Return dict mapping each non-secret storage record of interest by identifier or,
        for no filter specified, mapping them all. If wallet has no such item, return empty dict.

        :param typ: non-secret storage record type
        :param filt: non-secret storage record identifier or WQL json (default all)
        :param canon_wql: WQL canonicalization function (default von_anchor.canon.canon_non_secret_wql())
        :param limit: maximum number of results to return (default no limit)
        :return: dict mapping identifiers to non-secret storage records
        """

        LOGGER.debug('Wallet.get_non_secret >>> typ: %s, filt: %s, canon_wql: %s', typ, filt, canon_wql)

        if not self.handle:
            LOGGER.debug('Wallet.get_non_secret <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        records = []
        if isinstance(filt, str):  # ordinary lookup by value
            try:
                records = [json.loads(await non_secrets.get_wallet_record(
                    self.handle,
                    typ,
                    filt,
                    json.dumps({
                        'retrieveType': False,
                        'retrieveValue': True,
                        'retrieveTags': True
                    })))]
            except IndyError as x_indy:
                if x_indy.error_code == ErrorCode.WalletItemNotFound:
                    pass
                else:
                    LOGGER.debug(
                        'Wallet.get_non_secret <!< Wallet %s lookup raised indy exception %s',
                        self.name,
                        x_indy.error_code)
                    raise
        else:
            canon = canon_wql or canon_non_secret_wql
            s_handle = await non_secrets.open_wallet_search(
                self.handle,
                typ,
                json.dumps(canon(filt or {})),
                json.dumps({
                    'retrieveRecords': True,
                    'retrieveTotalCount': True,
                    'retrieveType': False,
                    'retrieveValue': True,
                    'retrieveTags': True
                }))

            records = []
            cardinality = int(json.loads(
                await non_secrets.fetch_wallet_search_next_records(self.handle, s_handle, 0))['totalCount'])
            chunk = min(cardinality, limit or cardinality, Wallet.DEFAULT_CHUNK)
            if limit:
                cardinality = min(limit, cardinality)
            try:
                while len(records) != cardinality:
                    batch = json.loads(
                        await non_secrets.fetch_wallet_search_next_records(self.handle, s_handle, chunk))['records']
                    records.extend(batch)
                    if len(batch) < chunk:
                        break
                if len(records) != cardinality:
                    LOGGER.warning(
                        'Non-secret search/limit indicated %s results but fetched %s',
                        cardinality,
                        len(records))
            finally:
                await non_secrets.close_wallet_search(s_handle)

        rv = {record['id']: StorageRecord(typ, record['value'], record['tags'], record['id']) for record in records}
        LOGGER.debug('Wallet.get_non_secret <<< %s', rv)
        return rv

    async def encrypt(
            self,
            message: bytes,
            authn: bool = False,
            to_verkey: str = None,
            from_verkey: str = None) -> bytes:
        """
        Encrypt plaintext for owner of DID, anonymously or via authenticated encryption scheme.
        Raise AbsentMessage for missing message, or WalletState if wallet is closed.

        :param message: plaintext, as bytes
        :param authn: whether to use authenticated encryption scheme
        :param to_verkey: verification key of recipient, None for anchor's own
        :param from_verkey: verification key of sender for authenticated encryption, None for anchor's own
        :return: ciphertext, as bytes
        """

        LOGGER.debug(
            'Wallet.encrypt >>> message: %s, authn: %s, to_verkey: %s, from_verkey: %s',
            message,
            authn,
            to_verkey,
            from_verkey)

        if not message:
            LOGGER.debug('Wallet.encrypt <!< No message to encrypt')
            raise AbsentMessage('No message to encrypt')

        if not self.handle:
            LOGGER.debug('Wallet.encrypt <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        if authn:
            rv = await crypto.auth_crypt(self.handle, from_verkey or self.verkey, to_verkey or self.verkey, message)
        else:
            rv = await crypto.anon_crypt(to_verkey or self.verkey, message)

        LOGGER.debug('Wallet.auth_encrypt <<< %s', rv)
        return rv

    async def decrypt(
            self,
            ciphertext: bytes,
            authn_check: bool = None,
            to_verkey: str = None,
            from_verkey: str = None) -> (bytes, str):
        """
        Decrypt ciphertext and optionally authenticate sender.

        Raise BadKey if authentication operation checks and reveals sender key distinct from input
        sender verification key.  Raise AbsentMessage for missing ciphertext, or WalletState if
        wallet is closed.

        :param ciphertext: ciphertext, as bytes
        :param authn_check: True to authenticate and check sender verification key,
            False to authenticate and return sender verification key for client to decide fitness, or
            None to use anonymous decryption
        :param to_verkey: recipient verification key, default anchor's own
        :param from_verkey: sender verification key, ignored for anonymous decryption,
            default anchor's own for authenticated decryption
        :return: decrypted bytes and sender verification key (None for anonymous decryption)
        """

        LOGGER.debug(
            'Wallet.decrypt >>> ciphertext: %s, authn_check: %s, to_verkey: %s, from_verkey: %s',
            ciphertext,
            authn_check,
            to_verkey,
            from_verkey)

        if not ciphertext:
            LOGGER.debug('Wallet.decrypt <!< No ciphertext to decrypt')
            raise AbsentMessage('No ciphertext to decrypt')

        if not self.handle:
            LOGGER.debug('Wallet.decrypt <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        sender_verkey = None
        if authn_check is None:
            plaintext = await crypto.anon_decrypt(self.handle, to_verkey or self.verkey, ciphertext)
        else:
            (sender_verkey, plaintext) = await crypto.auth_decrypt(self.handle, to_verkey or self.verkey, ciphertext)
            if authn_check and sender_verkey != (from_verkey or self.verkey):
                LOGGER.debug('Wallet.decrypt <!< Authentication revealed unexpected sender key on decryption')
                raise BadKey('Authentication revealed unexpected sender key on decryption')

        rv = (plaintext, sender_verkey)
        LOGGER.debug('Wallet.decrypt <<< %s', rv)
        return rv

    async def sign(self, message: bytes, verkey: str = None) -> bytes:
        """
        Derive signing key and Sign message; return signature. Raise WalletState if wallet is closed.
        Raise AbsentMessage for missing message, or WalletState if wallet is closed.

        :param message: Content to sign, as bytes
        :param verkey: verification key corresponding to private signing key (default anchor's own)
        :return: signature, as bytes
        """

        LOGGER.debug('Wallet.sign >>> message: %s, verkey: %s', message, verkey)

        if not message:
            LOGGER.debug('Wallet.sign <!< No message to sign')
            raise AbsentMessage('No message to sign')

        if not self.handle:
            LOGGER.debug('Wallet.sign <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = await crypto.crypto_sign(self.handle, verkey or self.verkey, message)

        LOGGER.debug('Wallet.sign <<< %s', rv)
        return rv

    async def verify(self, message: bytes, signature: bytes, verkey: str = None) -> bool:
        """
        Verify signature against input signer verification key (default anchor's own).
        Raise AbsentMessage for missing message or signature, or WalletState if wallet is closed.

        :param message: Content to sign, as bytes
        :param signature: signature, as bytes
        :param verkey: signer verification key (default for anchor's own)
        :return: whether signature is valid
        """

        LOGGER.debug('Wallet.verify >>> message: %s, signature: %s, verkey: %s', message, signature, verkey)

        if not message:
            LOGGER.debug('Wallet.verify <!< No message to verify')
            raise AbsentMessage('No message to verify')

        if not signature:
            LOGGER.debug('Wallet.verify <!< No signature to verify')
            raise AbsentMessage('No signature to verify')

        if not self.handle:
            LOGGER.debug('Wallet.verify <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = await crypto.crypto_verify(verkey or self.verkey, message, signature)

        LOGGER.debug('Wallet.verify <<< %s', rv)
        return rv

    async def pack(
            self,
            message: str,
            recip_verkeys: Union[str, Sequence[str]] = None,
            sender_verkey: str = None) -> bytes:
        """
        Pack a message for one or more recipients (default anchor only).
        Raise AbsentMessage for missing message, or WalletState if wallet is closed.

        :param message: message to pack
        :param recip_verkeys: verification keys of recipients (default anchor's own, only)
        :param sender_verkey: sender verification key (default anonymous encryption)
        :return: packed message
        """

        LOGGER.debug(
            'Wallet.pack >>> message: %s, recip_verkeys: %s, sender_verkey: %s',
            message,
            recip_verkeys,
            sender_verkey)

        if message is None:
            LOGGER.debug('Wallet.pack <!< No message to pack')
            raise AbsentMessage('No message to pack')

        rv = await crypto.pack_message(
            self.handle,
            message,
            [recip_verkeys] if isinstance(recip_verkeys, str) else list(recip_verkeys or [self.verkey]),
            sender_verkey)

        LOGGER.debug('Wallet.pack <<< %s', rv)
        return rv

    async def unpack(self, ciphertext: bytes) -> (str, str, str):
        """
        Unpack a message. Return triple with cleartext, sender verification key, and recipient verification key.
        Raise AbsentMessage for missing ciphertext, or WalletState if wallet is closed. Raise AbsentRecord
        if wallet has no key to unpack ciphertext.

        :param ciphertext: JWE-like formatted message as pack() produces
        :return: cleartext, sender verification key, recipient verification key
        """

        LOGGER.debug('Wallet.unpack >>> ciphertext: %s', ciphertext)

        if not ciphertext:
            LOGGER.debug('Wallet.pack <!< No ciphertext to unpack')
            raise AbsentMessage('No ciphertext to unpack')

        try:
            unpacked = json.loads(await crypto.unpack_message(self.handle, ciphertext))
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                LOGGER.debug('Wallet.unpack <!< Wallet %s has no local key to unpack ciphertext', self.name)
                raise AbsentRecord('Wallet {} has no local key to unpack ciphertext'.format(self.name))
            LOGGER.debug('Wallet.unpack <!< Wallet %s unpack() raised indy error code {}', x_indy.error_code)
            raise

        rv = (unpacked['message'], unpacked.get('sender_verkey', None), unpacked.get('recipient_verkey', None))
        LOGGER.debug('Wallet.unpack <<< %s', rv)
        return rv

    async def reseed_init(self, next_seed: str = None) -> str:
        """
        Begin reseed operation: generate new key. Raise WalletState if wallet is closed.

        :param next_seed: incoming replacement seed (default random)
        :return: new verification key
        """

        LOGGER.debug('Wallet.reseed_init >>> next_seed: [SEED]')

        if not self.handle:
            LOGGER.debug('Wallet.reseed_init <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        rv = await did.replace_keys_start(self.handle, self.did, json.dumps({'seed': next_seed} if next_seed else {}))
        LOGGER.debug('Wallet.reseed_init <<< %s', rv)
        return rv

    async def reseed_apply(self) -> DIDInfo:
        """
        Replace verification key with new verification key from reseed operation.
        Raise WalletState if wallet is closed.

        :return: DIDInfo with new verification key and metadata for DID
        """

        LOGGER.debug('Wallet.reseed_apply >>>')

        if not self.handle:
            LOGGER.debug('Wallet.reseed_init <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        await did.replace_keys_apply(self.handle, self.did)
        self.verkey = await did.key_for_local_did(self.handle, self.did)
        now = int(time())
        rv = DIDInfo(self.did, self.verkey, {'anchor': True, 'since': now, 'modified': now})
        await did.set_did_metadata(self.handle, self.did, json.dumps(rv.metadata))

        LOGGER.info('Wallet %s set seed hash metadata for DID %s', self.name, self.did)

        LOGGER.debug('Wallet.reseed_apply <<< %s', rv)
        return rv

    def __repr__(self) -> str:
        """
        Return representation for current object.

        :return: representation for current object
        """

        return 'Wallet({}, [ACCESS], {})'.format(self.config, self.auto_remove)
